'''Efficient container of properties for a reaction network.'''

from pySubnetSB import constants as cn  # type: ignore
from pySubnetSB.matrix import Matrix  # type: ignore
from pySubnetSB.named_matrix import NamedMatrix  # type: ignore
from pySubnetSB.stoichometry import Stoichiometry  # type: ignore
from pySubnetSB import util  # type: ignore
import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.assignment_pair import AssignmentPair  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint  # type: ignore
from pySubnetSB.species_constraint import SpeciesConstraint  # type: ignore

import collections
import json
import numpy as np
import os
import pandas as pd  # type: ignore
from pynauty import Graph  # type: ignore
import tellurium as te  # type: ignore
from typing import Optional, Tuple, Any, Union


Edge = collections.namedtuple('Edge', ['source', 'destination'])
GraphDescriptor = collections.namedtuple('GraphDescriptor', ['vertex_dct', 'label_dct'])
ConstraintPair = collections.namedtuple('ConstraintPair', ['reaction', 'species'])
ReferenceAndTargetResult = collections.namedtuple('ReferenceAndTargetResult',
      ['reference_network', 'target_network'])


class NetworkBase(object):
    """
    Abstraction for a reaction network. This is represented by reactant and product stoichiometry matrices.
    """

    def __init__(self, reactant_arr:Matrix, 
                 product_arr:np.ndarray,
                 reaction_names:Optional[np.ndarray[str]]=None,  # type: ignore
                 species_names:Optional[np.ndarray[str]]=None,   # type: ignore
                 network_name:Optional[str]=None)->None:
        """
        Args:
            reactant_arr (np.ndarray): Reactant matrix.
            product_arr (np.ndarray): Product matrix.
            network_name (str): Name of the network.
            reaction_names (np.ndarray[str]): Names of the reactions.
            species_names (np.ndarray[str]): Names of the species
        """
        self._network_name = network_name
        # Reactant stoichiometry matrix is negative
        if not np.all(reactant_arr.shape == product_arr.shape):
            raise ValueError("Reactant and product matrices must have the same shape.")
        try:
            self.num_species, self.num_reaction = np.shape(reactant_arr)
        except:
            # Empty network
            self.num_species = 0
            self.num_reaction = 0
            return
        if reactant_arr.shape[0] == 0:
            # Empty network
            self.num_species = 0
            self.num_reaction = 0
            return
        self.reactant_arr = reactant_arr
        self.product_arr = product_arr
        self.current_species_names = species_names
        self.current_reaction_names = reaction_names
        #
        self.reactant_nmat = NamedMatrix(self.reactant_arr,
              row_names=self.current_species_names, column_names=self.current_reaction_names,
              row_description="species", column_description="reactions")
        self.product_nmat = NamedMatrix(self.product_arr,
              row_names=self.current_species_names, column_names=self.current_reaction_names,
              row_description="species", column_description="reactions")
        self.standard_nmat = NamedMatrix(self.product_arr - self.reactant_arr,
              row_names=self.current_species_names,
              column_names=self.current_reaction_names, row_description="species", column_description="reactions")
        #
        self._species_names = self.current_species_names
        self._reaction_names = self.current_reaction_names
        self._stoichiometry_nmat:Optional[NamedMatrix] = None
        self._network_hash:Optional[int] = None  # Hash for weak identity
        self._constraint_pair_dct:dict = {}  # keys are identity, is_subnet

    def __bool__(self)->bool:
        return (self.num_species > 0) and (self.num_reaction > 0)

    @property
    def stoichiometry_nmat(self)->NamedMatrix:
        if self._stoichiometry_nmat is None:
            self._stoichiometry_nmat = NamedMatrix(self.product_nmat.values - self.reactant_nmat.values,
                  row_names=self.species_names, column_names=self.reaction_names,
                  row_description="species", column_description="reactions")
        return self._stoichiometry_nmat

    @property
    def species_names(self)->np.ndarray[str]:  # type: ignore
        if self._species_names is None:
            self._species_names = np.array([f"S{i}" for i in range(self.num_species)])
        if not isinstance(self._species_names, np.ndarray):
            self._species_names = np.array(self._species_names)
        return self._species_names
    
    @property
    def reaction_names(self)->np.ndarray[str]:  # type: ignore
        if self._reaction_names is None:
            self._reaction_names = np.array([f"J{i}" for i in range(self.num_reaction)])
        if not isinstance(self._reaction_names, np.ndarray):
            self._reaction_names = np.array(self._reaction_names)
        return self._reaction_names

    @property
    def network_hash(self):
        if self._network_hash is None:
            self._network_hash = util.hashMatrix(self.standard_nmat.values)
        return self._network_hash

    @property
    def network_name(self)->str:
        if self._network_name is None:
            self._network_name = str(np.random.randint(0, 10000000))
        return self._network_name
    
    def resetNetworkName(self)->None:
        self._network_name = None

    def copy(self)->'NetworkBase':
        return NetworkBase(self.reactant_nmat.values.copy(), self.product_nmat.values.copy(),
                       network_name=self.network_name, reaction_names=self.reaction_names,
                       species_names=self.species_names)

    def __repr__(self)->str:
        repr = f"{self.network_name}: {self.num_species} species, {self.num_reaction} reactions"
        reactions = ["  " + self.prettyPrintReaction(i) for i in range(self.num_reaction)]
        repr += '\n' + '\n'.join(reactions)
        return repr
    
    def getConstraints(self, identity:str, is_subnet:bool)->ConstraintPair:
        """
        Get the constraints for the network.

        Args:
            identity (str): cn.ID_STRONG or cn.ID_WEAK
            is_subnet (bool): True if the constraints are for a subset.

        Returns:
            Tuple[SpeciesConstraint, ReactionConstraint]
        """
        key = (identity, is_subnet)
        if not key in self._constraint_pair_dct:
            reactant_nmat, product_nmat = self.makeMatricesForIdentity(identity)
            constraint_pair = ConstraintPair(
                  ReactionConstraint(reactant_nmat, product_nmat, is_subnet=is_subnet),
                  SpeciesConstraint(reactant_nmat, product_nmat, is_subnet=is_subnet))
            self._constraint_pair_dct[key] = constraint_pair
        return self._constraint_pair_dct[key]
    
    def makeMatricesForIdentity(self, identity:str)->Tuple[NamedMatrix, NamedMatrix]:
        """Calculates the reactant and product NamedMatrix based on the identity calculation being done.

        Args:
            identity (str): cn.ID_STRONG or cn.ID_WEAK
        """
        if identity == cn.ID_STRONG:
            return self.reactant_nmat, self.product_nmat
        else:
            reactant_arr = -1*(self.standard_nmat.values<0)*self.standard_nmat.values
            reactant_nmat = self.reactant_nmat.copy()
            reactant_nmat.values = reactant_arr
            product_arr = (self.standard_nmat.values>0)*self.standard_nmat.values
            product_nmat = self.product_nmat.copy()
            product_nmat.values = product_arr
            return reactant_nmat, product_nmat
        
    def isBoundaryNetwork(self)->bool:
        """
        A boundary network is one where all reactions are either synthesis or degradation of a single species.

        Args:
            network (Network): Network instance

        Returns:
            bool: True if the network has only one species
        """
        reactant_sum_arr = self.reactant_nmat.values.sum(axis=0)
        product_sum_arr = self.product_nmat.values.sum(axis=0)
        is_boundary = np.all((reactant_sum_arr + product_sum_arr) <= 1)
        return bool(is_boundary)
    
    def isMatrixEqual(self, other, identity:str=cn.ID_WEAK)->bool:
        """
        Check if the stoichiometry matrix is equal to another network's matrix.
            weak identity: standard stoichiometry matrix 
            strong identity: reactant and product matrices

        Args:
            other: Network
            identity (str, optional): Defaults to cn.ID_WEAK.

        Returns:
            bool
        """
        reference_reactant_nmat, reference_product_nmat = self.makeMatricesForIdentity(identity)
        target_reactant_nmat, target_product_nmat = other.makeMatricesForIdentity(identity)
        return bool(np.all(reference_reactant_nmat.values == target_reactant_nmat.values) 
            and np.all(reference_product_nmat.values == target_product_nmat.values))
    
    def __eq__(self, other)->bool:
        if self.network_name != other.network_name:
            return False
        return self.isEquivalent(other)
    
    def isEquivalent(self, other)->bool:
        """Same except for the network name.

        Args:
            other (_type_): _description_

        Returns:
            bool: _description_
        """
        if not isinstance(other, self.__class__):
            return False
        if not self.isMatrixEqual(other, identity=cn.ID_STRONG):
            return False
        if not np.all(self.species_names == other.species_names):
            return False
        if not np.all(self.reaction_names == other.reaction_names):
            return False
        return True
    
    def permute(self, assignment_pair:Optional[AssignmentPair]=None)->Tuple['NetworkBase', AssignmentPair]:
        """
        Creates a new network with permuted reactant and product matrices. If no permutation is specified,
        then a random permutation is used.

        Returns:
            BaseNetwork (class of caller)
            AssignmentPair (species_assignment, reaction_assignment) for reconstructing the original network.
        """
        #####
        def makePerm(size:int)->np.ndarray[int]:  # type: ignore
            # Creates a permutation of the desired legnth, ensuring that it's not the identity permutation
            identity = np.array(range(size))
            if size == 1:
                return identity
            for _ in range(100):
                perm = np.random.permutation(range(size))
                if not np.all(perm == identity):
                    break
            else:
                raise RuntimeError("Could not find a permutation.")
            return perm
            #####
        if assignment_pair is None:
            reaction_perm = makePerm(self.num_reaction)
            species_perm = makePerm(self.num_species)
        else:
            reaction_perm = assignment_pair.reaction_assignment
            species_perm = assignment_pair.species_assignment
        reactant_arr = self.reactant_nmat.values.copy()
        product_arr = self.product_nmat.values.copy()
        reactant_arr = reactant_arr[species_perm, :]
        reactant_arr = reactant_arr[:, reaction_perm]
        product_arr = product_arr[species_perm, :]
        product_arr = product_arr[:, reaction_perm]
        reaction_names = np.array(self.reaction_names[reaction_perm])
        species_names = np.array(self.species_names[species_perm])
        assignment_pair = AssignmentPair(np.argsort(species_perm), np.argsort(reaction_perm))
        return self.__class__(reactant_arr, product_arr,
              reaction_names=reaction_names, species_names=species_names), assignment_pair
    
    def isStructurallyCompatible(self, other:'NetworkBase', identity:str=cn.ID_WEAK)->bool:
        """
        Determines if two networks are compatible to be structurally identical.
        This means that they have the same species and reactions.

        Args:
            other (Network): Network to compare to.
            identity (str): cn.ID_WEAK or cn.ID_STRONG

        Returns:
            bool: True if compatible.
        """
        if self.num_species != other.num_species:
            return False
        if self.num_reaction != other.num_reaction:
            return False
        is_identity = self.network_hash == other.network_hash
        if identity == cn.ID_STRONG:
            is_identity = self.network_hash == other.network_hash
        return bool(is_identity)

    # FIXME: More sophisticated subset checking? 
    def isSubsetCompatible(self, other:'NetworkBase')->bool:
        """
        Determines if two networks are compatible in that self can be a subset of other.
        This means that they have the same species and reactions.

        Args:
            other (Network): Network to compare to.

        Returns:
            bool: True if compatible.
        """
        if self.num_species > other.num_species:
            return False
        if self.num_reaction > other.num_reaction:
            return False
        return True
    
    @classmethod
    def makeFromAntimonyStr(cls, antimony_str:str, network_name:Optional[str]=None,
          roadrunner:Optional[Any]=None)->Union['NetworkBase', None]:
        """
        Make a Network from an Antimony string.

        Args:
            antimony_str (str): Antimony string.
            network_name (str): Name of the network.
            roadrunner (Any): Roadrunner object.

        Returns:
            Network or None
        """
        if roadrunner is None:
            clean_antimony_str = antimony_str.replace("\n", ";\n")
        else:
            clean_antimony_str = None
        stoichiometry = Stoichiometry(clean_antimony_str, roadrunner=roadrunner)
        if stoichiometry.reactant_mat is None:
            return None
        network = cls(stoichiometry.reactant_mat, stoichiometry.product_mat, network_name=network_name,
                      species_names=stoichiometry.species_names, reaction_names=stoichiometry.reaction_names)
        return network
                   
    @classmethod
    def makeFromAntimonyFile(cls, antimony_path:str,
                         network_name:Optional[str]=None)->Union['NetworkBase', None]:
        """
        Make a Network from an Antimony file. The default network name is the file name.

        Args:
            antimony_path (str): path to an Antimony file.
            network_name (str): Name of the network.

        Returns:
            Network
        """
        with open(antimony_path, 'r') as fd:
            try:
                antimony_str = fd.read()
            except:
                raise ValueError(f"Could not read file {antimony_path}")
        if network_name is None:
            filename = os.path.basename(antimony_path)
            network_name = filename.split('.')[0]
        try:
            network = cls.makeFromAntimonyStr(antimony_str, network_name=network_name)
        except:
            network = None
        return network
    
    @classmethod
    def makeFromSBMLFile(cls, sbml_path:str,
                         network_name:Optional[str]=None)->Union['NetworkBase', None]:
        """
        Make a Network from an SBML file. The default network name is the file name.

        Args:
            sbml_path (str): path to an SBML file.
            network_name (str): Name of the network.

        Returns:
            Network
        """
        roadrunner = te.loadSBMLModel(sbml_path)
        stoichiometry = Stoichiometry(None, roadrunner=roadrunner)
        if stoichiometry.reactant_mat is None:
            return None
        network = cls(stoichiometry.reactant_mat, stoichiometry.product_mat, network_name=network_name,
                      species_names=stoichiometry.species_names, reaction_names=stoichiometry.reaction_names)
        return network
    
    @classmethod
    def makeRandomNetwork(cls, num_species:int=5, num_reaction:int=5,
          )->'NetworkBase':
        """
        Makes a random network.

        Args:
            num_species (int): Number of species.
            num_reaction (int): Number of reactions.

        Returns:
            Network
        """
        reactant_arr = np.random.randint(0, 3, (num_species, num_reaction))
        product_arr = np.random.randint(0, 3, (num_species, num_reaction))
        return cls(reactant_arr, product_arr)
    
    @classmethod
    def makeRandomNetworkByReactionType(cls, num_reaction:int, num_species:int=-1, is_exact:bool=False,
              species_names:Optional[np.ndarray]=None, reaction_names:Optional[np.ndarray]=None,
              **kwargs)->'NetworkBase':
        """
        Makes a random network based on the type of reaction. Parameers are in the form
            <p<#products>r<#reactants>_frc> where #products and #reactants are the number of
            products and reactants. Ensures that reactions are unique.
        Fractions are from the paper "SBMLKinetics: a tool for annotation-independent classification of
            reaction kinetics for SBML models", Jin Liu, BMC Bioinformatics, 2023.

        Args:
            num_reaction (int): Number of reactions.
            num_species (int): Number of species. If <0, then num_species = num_reaction.
            is_exact (bool): Ensure that the number of species is exactly num_species.
            is_unique (bool): Ensure that reactions are unique.
            is_prune_species (bool): Prune species not used in any reaction.
            fractions by number of products and reactants
            species_names (np.ndarray): Names of the species
            reaction_names (np.ndarray): Names of the reactions

        Returns:
            Network
        """
        if num_reaction <= 0:
            raise ValueError("Number of reactions must be positive.")
        if not is_exact:
            return cls._makeRandomNetworkByReactionType(num_reaction, num_species=num_species,
                  species_names=species_names, reaction_names=reaction_names, **kwargs)
        # Get the network consistent with the requirements
        max_attempts =  100000
        if num_species < 0:
            num_species = num_reaction
        for _ in range(max_attempts): 
            network = cls._makeRandomNetworkByReactionType(num_reaction, num_species=num_species,
                  species_names=species_names, reaction_names=reaction_names, **kwargs)
            if network.num_species == num_species:
                return network
        raise ValueError(
              f"Could not find a network with {num_species} species and {num_reaction} reactions.")

    
    @classmethod
    def _makeRandomNetworkByReactionType(cls, 
              num_reaction:int,
              num_species:int=-1,
              species_names:Optional[np.ndarray]=None,
              reaction_names:Optional[np.ndarray]=None,
              is_prune_species:bool=True,
              p0r0_frc:Optional[float]=0.0,
              p0r1_frc:Optional[float]=0.1358,
              p0r2_frc:Optional[float]=0.001,
              p0r3_frc:Optional[float]=0.0,
              p1r0_frc:Optional[float]=0.0978,
              p1r1_frc:Optional[float]=0.3364,
              p1r2_frc:Optional[float]=0.1874,
              p1r3_frc:Optional[float]=0.0011,
              p2r0_frc:Optional[float]=0.0005,
              p2r1_frc:Optional[float]=0.1275,
              p2r2_frc:Optional[float]=0.0683,
              p2r3_frc:Optional[float]=0.0055,
              p3r0_frc:Optional[float]=0.0,
              p3r1_frc:Optional[float]=0.0087,
              p3r2_frc:Optional[float]=0.0154,
              p3r3_frc:Optional[float]=0.0146,
              is_unique:bool=False,
              )->'NetworkBase':
        """
        Makes a random network based on the type of reaction. Parameers are in the form
            <p<#products>r<#reactants>_frc> where #products and #reactants are the number of
            products and reactants. Ensures that reactions are unique.
        Fractions are from the paper "SBMLKinetics: a tool for annotation-independent classification of
            reaction kinetics for SBML models", Jin Liu, BMC Bioinformatics, 2023.

        Args:
            num_reaction (int): Number of reactions.
            num_species (int): Number of species.
            is_unique (bool): Ensure that reactions are unique.
            is_prune_species (bool): Prune species not used in any reaction.
            fractions by number of products and reactants
            species_names (np.ndarray): Names of the species
            reaction_names (np.ndarray): Names of the reactions

        Returns:
            Network
        """
        if reaction_names is not None:
            num_reaction = len(reaction_names)
        if species_names is not None:
            num_species = len(species_names)
        if num_species < 0:
            num_species = num_reaction
        SUFFIX = "_frc"
        # Initializations
        REACTION_TYPES = [f"p{i}r{j}" for i in range(4) for j in range(4)]
        FRAC_NAMES = [n + SUFFIX for n in REACTION_TYPES]
        value_dct:dict = {}
        total = 0
        for name in FRAC_NAMES:
            total += locals()[name] if locals()[name] is not None else 0
        for name in FRAC_NAMES:
            value = locals()[name]
            value_dct[name] = value/total
        CULMULATIVE_ARR = np.cumsum([value_dct[n + SUFFIX] for n in REACTION_TYPES])
        #######
        def getType(frac:float)->str:
            """
            Returns the name of the reaction associated with the fraction (e.g., a random (0, 1))

            Args:
                frac (float)

            Returns:
                str: Reaction type
            """
            pos = np.sum(CULMULATIVE_ARR < frac)
            for _ in range(len(REACTION_TYPES)):
                reaction_type = REACTION_TYPES[pos]
                # Handle cases of 0 fractions
                if not np.isclose(value_dct[reaction_type + SUFFIX], frac):
                    break
            return reaction_type
        #######
        MAX_ITERATION = 1000
        # Construct the reactions by building the reactant and product matrices
        reactant_arr = np.zeros((num_species, num_reaction))
        product_arr = np.zeros((num_species, num_reaction))
        i_reaction = 0
        num_iteration = 0
        while i_reaction < num_reaction:
            num_iteration += 1
            if num_iteration > max(num_reaction, MAX_ITERATION):
                raise ValueError("Could not find unique reactions.")
            frac = np.random.rand()
            reaction_type = getType(frac)
            num_product = int(reaction_type[1])
            num_reactant = int(reaction_type[3])
            # Products
            product_idxs = np.random.randint(0, num_species, num_product)
            product_arr[product_idxs, i_reaction] += 1
            # Reactants
            reactant_idxs = np.random.randint(0, num_species, num_reactant)
            reactant_arr[reactant_idxs, i_reaction] += 1
            # Check for unique reactions
            merged_arr = np.hstack([reactant_arr.T[:i_reaction+1], product_arr.T[:i_reaction+1]])
            if is_unique:
                uniques = np.unique(merged_arr, axis=0)
                if len(uniques) == i_reaction + 1:
                    # Unique reaction. Keep it
                    i_reaction += 1
                else:
                    # Duplicated reaction. Back it out
                    product_arr[product_idxs, i_reaction] -= 1
                    reactant_arr[reactant_idxs, i_reaction] -= 1
            else:
                i_reaction += 1
        # Eliminate 0 rows (species not used)
        if is_prune_species:
            keep_idxs:list = []
            for i_species in range(num_species):
                if np.sum(reactant_arr[i_species, :]) > 0 or np.sum(product_arr[i_species, :]) > 0:
                    keep_idxs.append(i_species)
            reactant_arr = reactant_arr[keep_idxs, :]
            product_arr = product_arr[keep_idxs, :]
        # Construct the network
        actual_num_species = reactant_arr.shape[0]
        actual_num_reaction = reactant_arr.shape[1]
        if species_names is None:
            species_names = util.getDefaultSpeciesNames(actual_num_species)
        else:
            species_names = np.array(species_names[:actual_num_species])
        if reaction_names is None:
            reaction_names = util.getDefaultReactionNames(actual_num_reaction)
        else:
            reaction_names = np.array(reaction_names[:actual_num_reaction])
        network = cls(reactant_arr, product_arr, reaction_names=reaction_names, species_names=species_names)
        return network
   
    def prettyPrintReaction(self, index:int)->str:
        """
        Pretty prints a reaction.

        Args:
            index (int): Index of the reaction.

        Returns:
            str
        """
        def makeSpeciesExpression(reaction_idx:int, stoichiometry_arr:np.ndarray)->str:
            all_idxs = np.array(range(self.num_species))
            species_idxs = all_idxs[stoichiometry_arr[:, reaction_idx] > 0]
            species_names = self.species_names[species_idxs]
            stoichiometries = [s for s in stoichiometry_arr[species_idxs, reaction_idx]]
            stoichiometries = ["" if np.isclose(s, 1) else str(s) + " " for s in stoichiometries]
            expressions = [f"{stoichiometries[i]}{species_names[i]}" for i in range(len(species_names))]
            result =  ' + '.join(expressions)
            return result
        #
        reactant_expression = makeSpeciesExpression(index, self.reactant_nmat.values)
        product_expression = makeSpeciesExpression(index, self.product_nmat.values)
        result = f"{self.reaction_names[index]}: " + f"{reactant_expression} -> {product_expression}"
        return result

    def makeNetworkFromAssignmentPair(self, assignment_pair:AssignmentPair)->'NetworkBase':
        """
        Constructs a network from an assignment pair.

        Args:
            assignment_pair (AssignmentPair): Assignment pair.

        Returns:
            Network: Network constructed from the assignment pair.
        """
        species_assignment = assignment_pair.species_assignment
        reaction_assignment = assignment_pair.reaction_assignment
        reactant_arr = self.reactant_nmat.values[species_assignment, :]
        product_arr = self.product_nmat.values[species_assignment, :]
        reactant_arr = reactant_arr[:, reaction_assignment]
        product_arr = product_arr[:, reaction_assignment]
        return NetworkBase(reactant_arr, product_arr, reaction_names=self.reaction_names[reaction_assignment],
                        species_names=self.species_names[species_assignment])
    
    def serialize(self)->str:
        """
        Serialize the network.

        Returns:
            str: string representation of json structure
        """
        reactant_lst = self.reactant_nmat.values.tolist()
        product_lst = self.product_nmat.values.tolist()
        dct = {cn.S_ID: self.__class__.__name__,
               cn.S_NETWORK_NAME: self.network_name,
               cn.S_REACTANT_LST: reactant_lst,
               cn.S_PRODUCT_LST: product_lst,
               cn.S_REACTION_NAMES: self.reaction_names.tolist(),
               cn.S_SPECIES_NAMES: self.species_names.tolist(),
               }
        return json.dumps(dct)

    @classmethod 
    def seriesToJson(cls, series:pd.Series)->str:
        """
        Convert a Series to a JSON serialization string.

        Args:
            ser: Series columns
                network_name, reactant_array_str, product_array_str, species_names, reaction_names,
                num_speces, um_reaction

        Returns:
            str: string representation of json structure
        """
        def convert(name):
            values = series[name]
            if isinstance(values, str):
                values = eval(values)
            return list(values)
        #####
        reactant_names = convert(cn.S_REACTION_NAMES)
        species_names = convert(cn.S_SPECIES_NAMES)
        num_reaction = len(reactant_names)
        num_species = len(species_names)
        product_arr = np.array(convert(cn.S_PRODUCT_LST))
        reactant_arr = np.array(convert(cn.S_REACTANT_LST))
        reactant_arr = np.reshape(reactant_arr, (num_species, num_reaction))
        product_arr = np.reshape(product_arr, (num_species, num_reaction))
        dct = {cn.S_ID: str(cls),
               cn.S_NETWORK_NAME: series[cn.S_NETWORK_NAME],
               cn.S_REACTANT_LST: reactant_arr.tolist(),
               cn.S_PRODUCT_LST: product_arr.tolist(),
               cn.S_REACTION_NAMES: reactant_names,
               cn.S_SPECIES_NAMES: species_names,
               }
        return json.dumps(dct)

    def toSeries(self)->pd.Series:
        """
        Serialize the network.

        Args:
            ser: Series columns
                network_name, reactant_array_str, product_array_str, species_names, reaction_names,
                num_speces, um_reaction

        Returns:
            str: string representation of json structure
        """
        dct = {cn.S_REACTANT_LST: self.reactant_nmat.values.flatten().tolist(),
               cn.S_PRODUCT_LST: self.product_nmat.values.flatten().tolist(),
               cn.S_NETWORK_NAME: self.network_name,
               cn.S_REACTION_NAMES: self.reaction_names,
               cn.S_SPECIES_NAMES: self.species_names}
        return pd.Series(dct)

    @classmethod 
    def deserialize(cls, serialization_str)->'NetworkBase':
        """
        Serialize the network.

        Returns:
            str: string representation of json structure
        """
        dct = json.loads(serialization_str)
        if not cls.__name__ in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls} but got {dct[cn.S_ID]}")
        network_name = dct[cn.S_NETWORK_NAME]
        reactant_arr = np.array(dct[cn.S_REACTANT_LST])
        product_arr = np.array(dct[cn.S_PRODUCT_LST])
        reaction_names = np.array(dct[cn.S_REACTION_NAMES])
        species_names = np.array(dct[cn.S_SPECIES_NAMES])
        return cls(reactant_arr, product_arr, network_name=network_name,
                       reaction_names=reaction_names, species_names=species_names)
    
    def getGraphDescriptor(self, identity=cn.ID_STRONG)->GraphDescriptor:
        """
        Describes the bipartite graph of the network.
        Species are indices 0 to num_species - 1 and reactions are
        indices num_species to num_species + num_reaction - 1.
        Labels are provided to designate reaction and species nodes.

        Args:
            identity (str): cn.ID_STRONG or cn.ID_WEAK

        Returns:
            vertex_dict:
                key: source vertex index
                value: list of detination indices
            label_dict:
                key: vertex
                value: label ('reaction' or 'species')
        """
        num_species = self.num_species
        num_reaction = self.num_reaction
        num_node = num_species + num_reaction
        # Vertex dictionary
        vertex_dct:dict = {i: [] for i in range(num_node)}
        for i_reaction in range(num_reaction):
            for i_species in range(num_species):
                # Add an edge for each reactant
                species_vtx = i_species
                reaction_vtx = num_species + i_reaction
                if identity == cn.ID_STRONG:
                    # Reactants
                    for _ in range(int(self.reactant_nmat.values[i_species, i_reaction])):
                        vertex_dct[species_vtx].append(reaction_vtx)
                    # Products
                    for _ in range(int(self.product_nmat.values[i_species, i_reaction])):
                        vertex_dct[reaction_vtx].append(species_vtx)
                else:
                    # Weak identity. Use the standard stoichiometry matrix
                    stoichiometry = int(self.standard_nmat.values[i_species, i_reaction])
                    for _ in range(np.abs(stoichiometry)):
                        if stoichiometry > 0:
                            # Product
                            vertex_dct[reaction_vtx].append(species_vtx)
                        elif stoichiometry < 0:
                            # Reactant
                            vertex_dct[species_vtx].append(reaction_vtx)
                        else:
                            pass
        # Label dictionary
        label_dct:dict = {n: 'species' for n in range(num_species)}
        label_dct.update({n: 'reaction' for n in range(num_species, num_node)})

        return GraphDescriptor(vertex_dct=vertex_dct, label_dct=label_dct)

    def makePynautyNetwork(self, identity=cn.ID_STRONG)->Graph:
        """
        Make a pynauty graph from the network. Species are indices 0 to num_species - 1 and reactions are
        indices num_species to num_species + num_reaction - 1.

        Args:
            identity (str): cn.ID_STRONG or cn.ID_WEAK

        Returns:
            Graph: Pynauty graph
        """
        graph_descriptor = self.getGraphDescriptor(identity=identity)
        vertex_dct = graph_descriptor.vertex_dct
        graph = Graph(len(vertex_dct.keys()), directed=True)
        for node, neighbors in vertex_dct.items():
            graph.connect_vertex(node, neighbors)
        return graph
    
    def makeCSVNetwork(self, identity=cn.ID_STRONG)->str:
        """
        Creates a CSV representation of a directed graph. (See https://github.com/ciaranm/glasgow-subgraph-solver)
        indices num_species to num_species + num_reaction - 1.
        The CSV format has one line for each edge.  j
          <source>> <destination>

        Args:
            identity (str): cn.ID_STRONG or cn.ID_WEAK

        Returns:
            Graph: Pynauty graph
        """
        graph_descriptor = self.getGraphDescriptor(identity=identity)
        outputs = []
        # Create the vertices
        for source, destinations in graph_descriptor.vertex_dct.items():
            for destination in destinations:
                outputs.append(f"{source}>{destination}")
        # Add the labels
        for vertex, label in graph_descriptor.label_dct.items():
            outputs.append(f"{vertex},,{label}")
        return '\n'.join(outputs)
    
    def fill(self, num_fill_species:int=3, num_fill_reaction:Optional[int]=3,
               is_permute:bool=True)->'NetworkBase':
        """Creates a new network that augments (fills) the existing network with new species and reactions.

        Args:
            num_fill_species (int): Number of filler species. Use num_reaction and num_species if None.
            num_fill_reaction (int): Number of filler reactions. Use num_species and num_reaction if None.
            is_permute (bool): Permute the network.

        Returns:
            Network
        """
        # Check options
        if num_fill_species is None and num_fill_reaction is None:
            raise ValueError("Must specify either num_fill_species or num_fill_reaction.")
        if num_fill_species is None:
            num_fill_species = num_fill_reaction
        if num_fill_reaction is None:
            num_fill_reaction = num_fill_species
        # Create the filler network
        if num_fill_species < 1:  # type: ignore
            raise ValueError("Number of filler species must be at least 1.")
        if num_fill_reaction < 1: # type: ignore
            raise ValueError("Number of filler reaction must be at least 1.")
        filler_network = NetworkBase.makeRandomNetworkByReactionType(num_fill_reaction,  # type: ignore
            num_species=num_fill_species)
        # Creates a supernetwork with the reference in the upper left corner of the matrices
        # and the filler network in the bottom right. Then, randomize.
        self_pad_arr = np.zeros((self.num_species, filler_network.num_reaction))   # type: ignore
        filler_pad_arr = np.zeros((filler_network.num_species, self.num_reaction)) # type: ignore
        #####
        def makeTargetArray(reference_arr:np.ndarray, filler_arr:np.ndarray)->np.ndarray:
            padded_reference_arr = np.hstack([reference_arr.copy(), self_pad_arr])
            padded_filler_arr = np.hstack([filler_pad_arr, filler_arr.copy()])
            target_arr = np.vstack([padded_reference_arr, padded_filler_arr])
            return target_arr
        ##### 
        # Construct the reactant array so that reference is upper left and filler is lower right
        target_reactant_arr = makeTargetArray(self.reactant_nmat.values,
              filler_network.reactant_nmat.values)
        target_product_arr = makeTargetArray(self.product_nmat.values,
              filler_network.product_nmat.values)
        target_network = self.__class__(target_reactant_arr, target_product_arr)
        if is_permute:
            target_network, _ = target_network.permute()
        return target_network

    # TODO: Optionally provided the reference network. Handling ensuring that target has the reference names.
    @classmethod
    def makeRandomReferenceAndTarget(cls,
              num_reference_species:int=3,
              num_target_species:int=6,
              num_reference_reaction:Optional[int]=None,
              num_target_reaction:Optional[int]=None,
              )->ReferenceAndTargetResult:
        """
        Creates a random reference and target network of the desired size
        with the reference network as a subnet of the target. The reference uses the species in the target.

        Args:
            num_reference_species (int): Number of species in the reference network.
            num_target_species (int): Number of species in the target network.
            num_reference_reaction (int): Number of reactions in the reference network.
            num_target_reaction (int): Number of reactions in the target network.
            species_names (np.ndarray): Species names.

        Returns:
            ReferenceAndTargetResult: Reference Network, Target Network
        """
        # Check options
        if num_reference_reaction is None:
            num_reference_reaction = num_reference_species
        if num_target_reaction is None:
            num_target_reaction = num_target_species
        # Error checks
        if num_reference_species > num_target_species:
            raise ValueError("Number of reference species must be less than or equal to the number of target species.")
        if num_reference_reaction >= num_target_reaction:
            raise ValueError("Number of reference reactions must be less than the number of target reactions.")
        # Create the networks.
        if num_target_reaction < 10:
            # Can do exact for smaller networks
            is_exact = True
        else:
            is_exact = False
        partial_target_network = cls.makeRandomNetworkByReactionType(num_target_reaction-num_reference_reaction,
            num_species=num_target_species, is_exact=is_exact)
        reference_species_name_arr = np.random.choice(partial_target_network.species_names, num_reference_species,
              replace=False)
        reference_reaction_name_arr = np.array(["J" + n for n in partial_target_network.reaction_names[:num_reference_reaction]])
        reference_network = cls.makeRandomNetworkByReactionType(num_reference_reaction,
             species_names=reference_species_name_arr, reaction_names=reference_reaction_name_arr,
             is_exact=is_exact)
        # Merge the stoichiometry matrices of the reference network with the target
        target_reactant_nmat = partial_target_network.reactant_nmat.vmerge(reference_network.reactant_nmat)
        target_product_nmat = partial_target_network.product_nmat.vmerge(reference_network.product_nmat)
        # Randomize the rows and columns of the target
        randomized_target_reactant_nmat_result = target_reactant_nmat.randomize()
        randomized_target_reactant_nmat = randomized_target_reactant_nmat_result.named_matrix
        randomized_target_product_nmat = target_product_nmat.randomize(
              row_perm=randomized_target_reactant_nmat_result.row_perm,
              column_perm=randomized_target_reactant_nmat_result.column_perm).named_matrix
        reaction_names = randomized_target_reactant_nmat.column_names
        species_names = randomized_target_reactant_nmat.row_names
        """ reaction_names = np.concatenate([partial_target_network.reaction_names, reference_network.reaction_names])
        reaction_names = reaction_names[randomized_target_reactant_nmat_result.column_perm]
        species_names = partial_target_network.species_names[randomized_target_reactant_nmat_result.row_perm] """
        randomized_target_network = cls(randomized_target_reactant_nmat.values, randomized_target_product_nmat.values,
                reaction_names=reaction_names, species_names=species_names)
        #
        return ReferenceAndTargetResult(
            reference_network=reference_network, 
            target_network=randomized_target_network
        )

    
    def makeInducedNetwork(self, assignment_pair:AssignmentPair)->'NetworkBase':
        """
        Makes an induced network based on an assignment pair.

        Args:
            assignment_pair (AssignmentPair): Assignment pair.

        Returns:
            Network
        """
        species_assignment = assignment_pair.species_assignment
        reaction_assignment = assignment_pair.reaction_assignment
        reactant_arr = self.reactant_nmat.values[species_assignment, :]
        product_arr = self.product_nmat.values[species_assignment, :]
        reactant_arr = reactant_arr[:, reaction_assignment]
        product_arr = product_arr[:, reaction_assignment]
        return self.__class__(reactant_arr, product_arr, reaction_names=self.reaction_names[reaction_assignment],
                        species_names=self.species_names[species_assignment])


from pySubnetSB.constraint import Constraint, ReactionClassification, CompatibilityCollection    # type: ignore
from pySubnetSB.named_matrix import NamedMatrix   # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint  # type: ignore
from pySubnetSB.constraint_option_collection import ReactionConstraintOptionCollection  # type: ignore

import itertools
import numpy as np
from scipy.special import factorial  # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
reactant_arr = np.array([[1, 0], [0, 1], [0, 0]]) # Must have 3 rows to be consistent with DummyConstraint
product_arr = np.array([[0, 1], [1, 0], [0, 0]])
NUM_SPECIES, NUM_REACTION = reactant_arr.shape
SPECIES_NAMES = ["S" + str(i) for i in range(NUM_SPECIES)]
REACTION_NAMES = ["J" + str(i) for i in range(NUM_REACTION)]
REACTANT_NMAT = NamedMatrix(reactant_arr,  row_names=SPECIES_NAMES, column_names=REACTION_NAMES)
PRODUCT_NMAT = NamedMatrix(product_arr,  row_names=SPECIES_NAMES, column_names=REACTION_NAMES)
MODEL1 = """
J0: S0 -> S1; k0*S0
J1: S1 -> S0; k1*S1
k0 = 0.1; k1 = 0.2;
S0 = 10; S1 = 20
"""
NETWORK1 = Network.makeFromAntimonyStr(MODEL1)
MODEL2 = """
J0: S0 -> S1; k0*S0
J1: S1 -> S1; k1*S1
J2: S0 + S1 -> S1; k2*S0*S1
k0 = 0.1; k1 = 0.2; k2 = 0.3
S0 = 10; S1 = 20; S2 = 30
"""
NETWORK2 = Network.makeFromAntimonyStr(MODEL2)

#############################
class DummyConstraint(Constraint):
    SPECIES_NAMES = ["A", "B", "C"]
    NUM_ROW = len(SPECIES_NAMES)
    arr = np.random.randint(0, 2, (NUM_ROW, NUM_ROW))
    ONE_STEP_NMAT = NamedMatrix(arr, row_names='rows', column_names='columns')

    
    def __init__(self, reactant_nmat:NamedMatrix, product_nmat:NamedMatrix):
        self.species_names = self.SPECIES_NAMES
        self._num_row = self.NUM_ROW
        super().__init__(reactant_nmat=reactant_nmat, product_nmat=product_nmat)
        #
        self._categorical_nmat = NamedMatrix(np.array([[0, 1], [1, 0], [1, 1]]),
                row_names=self.species_names, column_names=REACTION_NAMES)
        self._enumerated_nmat = NamedMatrix(np.array([[0, 10], [10, 0], [10, 10]]),
                row_names=self.species_names, column_names=REACTION_NAMES)
        
    def scale(self, scale:int)->'DummyConstraint':
        num_row = scale*len(self.species_names)
        reactant_nmat = NamedMatrix.vstack([self.reactant_nmat]*scale, is_rename=True)
        product_nmat = NamedMatrix.vstack([self.product_nmat]*scale, is_rename=True)
        constraint = DummyConstraint(reactant_nmat, product_nmat)
        equality_arr = np.vstack([self.equality_nmat.values]*scale)
        inequality_arr = np.vstack([self.numerical_inequality_nmat.values]*scale)
        constraint._categorical_nmat = NamedMatrix(equality_arr, row_names=np.array(range(num_row)),
                column_names=REACTION_NAMES)
        constraint._enumerated_nmat = NamedMatrix(inequality_arr, row_names=np.array(range(num_row)),
                column_names=REACTION_NAMES)
        return constraint

    @property
    def categorical_nmat(self)->NamedMatrix:
        return self._categorical_nmat
    
    @property
    def numerical_enumerated_nmat(self)->NamedMatrix:
        return self._enumerated_nmat

    @property
    def bitwise_enumerated_nmat(self)->NamedMatrix:
        return self._enumerated_nmat
    
    @property
    def one_step_nmat(self)->NamedMatrix:
        arr = np.matmul(self.product_nmat.values, self.reactant_nmat.values.T)
        row_names = [str(n) for n in range(arr.shape[0])]
        return NamedMatrix(arr, row_names=row_names, column_names=row_names)

    @classmethod 
    def makeDummyConstraint(cls, num_species:int=3, num_reaction:int=3):
        for _ in range(100):
            network = Network.makeRandomNetworkByReactionType(num_reaction, num_species)
            if (num_species != network.num_species) or (num_reaction != network.num_reaction):
                continue
            break
        else:
            raise RuntimeError("Failed to make a random network")
        return DummyConstraint(network.reactant_nmat, network.product_nmat)
    
    @property
    def row_names(self)->np.ndarray:
        return np.array(SPECIES_NAMES)


#############################
class ScalableDummyConstraint(Constraint):
    
    def __init__(self, reference_size):
        reactant_nmat = NamedMatrix(np.array(np.random.randint(0, 2, (reference_size, reference_size))))
        product_nmat = NamedMatrix(np.array(np.random.randint(0, 2, (reference_size, reference_size))))
        super().__init__(reactant_nmat=reactant_nmat, product_nmat=product_nmat)
        #
        self.num_species = len(reactant_nmat.row_names)
        num_column = 5
        self._categorical_nmat = NamedMatrix(np.array(np.random.randint(0, 2, (self.num_species, num_column))))
        self._enumerated_nmat = NamedMatrix(np.array(np.random.randint(0, 2, (self.num_species, num_column))))
        
    def scale(self, scale:int)->'ScalableDummyConstraint':
        reactant_nmat = NamedMatrix.vstack([self.reactant_nmat]*scale, is_rename=True)
        product_nmat = NamedMatrix.vstack([self.product_nmat]*scale, is_rename=True)
        constraint = DummyConstraint(reactant_nmat, product_nmat)
        equality_arr = np.vstack([self.equality_nmat.values]*scale)
        inequality_arr = np.vstack([self.numerical_inequality_nmat.values]*scale)
        num_row = equality_arr.shape[0]
        constraint._categorical_nmat = NamedMatrix(equality_arr, row_names=np.array(range(num_row)))
        constraint._enumerated_nmat = NamedMatrix(inequality_arr, row_names=np.array(range(num_row)))
        return constraint

    @property
    def categorical_nmat(self)->NamedMatrix:
        return self._categorical_nmat
    
    @property
    def numerical_enumerated_nmat(self)->NamedMatrix:
        return self._enumerated_nmat
    
    @property
    def bitwise_enumerated_nmat(self)->NamedMatrix:
        return self._enumerated_nmat
    
    @property
    def one_step_nmat(self)->NamedMatrix:
        arr = np.matmul(self.product_nmat.values, self.reactant_nmat.values.T)
        return NamedMatrix(arr,
                row_names='rows', column_names='columns')
    

#############################
class TestReactionClassification(unittest.TestCase):

    def setUp(self):
        self.reaction_classification = ReactionClassification(num_reactant=1, num_product=2)
    
    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(str(self.reaction_classification), "uni-bi")

    def testMany(self):
        if IGNORE_TEST:
            return
        iter =  itertools.product(range(4), repeat=2)
        for num_reactant, num_product in iter:
            reaction_classification = ReactionClassification(num_reactant=num_reactant,
                  num_product=num_product)
            if (num_reactant == 0) or (num_product == 0):
                self.assertTrue("null" in str(reaction_classification))
            if (num_reactant == 1) or (num_product == 1):
                self.assertTrue("uni" in str(reaction_classification))
            if (num_reactant == 2) or (num_product == 2):
                self.assertTrue("bi" in str(reaction_classification))
            if (num_reactant == 3) or (num_product == 3):
                self.assertTrue("multi" in str(reaction_classification))

    def testMakeReactionClassificationMatrix(self):
        if IGNORE_TEST:
            return
        reaction_classifications = [ReactionClassification(num_reactant=1, num_product=2),
              ReactionClassification(num_reactant=2, num_product=1)]
        reaction_names = ["J1", "J2"]
        nmat = ReactionClassification.makeReactionClassificationMatrix(reaction_names, reaction_classifications)
        df = nmat.dataframe
        self.assertTrue(df.loc["J1", "uni-bi"] == 1)
        self.assertTrue(df.loc["J2", "bi-uni"] == 1)
        for _, row in df.iterrows():
            self.assertEqual(np.sum(row), 1)


#############################
class TestCompatibilityCollection(unittest.TestCase):

    def setUp(self) -> None:
        self.collection = CompatibilityCollection(2, 3)

    def testNumPermutation(self):
        if IGNORE_TEST:
            return
        for size in np.random.randint(2, 100, 10):
            collection = CompatibilityCollection(size, size)
            [collection.add(i-1, range(i)) for i in range(1, size+1)]
            self.assertLessEqual(collection.log10_num_assignment, np.log10(factorial(size)))

    def testPrune(self):
        if IGNORE_TEST:
            return
        log10_max_permutation = 4.0
        for size in np.random.randint(5, 30, 100):
            collection = CompatibilityCollection(size, size)
            [collection.add(i-1, range(i)) for i in range(1, size+1)]
            new_collection, is_changed = collection.prune(log10_max_permutation)
            result = (collection.log10_num_assignment <= log10_max_permutation) and (not is_changed)
            result = result or (collection.log10_num_assignment > log10_max_permutation) and is_changed
            self.assertTrue(result)
            if not is_changed:
                self.assertEqual(new_collection, collection)
            else:
                self.assertNotEqual(new_collection, collection)

    def testExpand(self):
        if IGNORE_TEST:
            return
        size = 3
        collection = CompatibilityCollection(size, size)
        [collection.add(i-1, range(i)) for i in range(1, size+1)]
        arr, _ = collection.expand()
        self.assertLessEqual(arr.shape[0], factorial(3))
        self.assertEqual(arr.shape[1], 3)


#############################
class TestConstraint(unittest.TestCase):

    def setUp(self):
        self.constraint = DummyConstraint(reactant_nmat=REACTANT_NMAT, product_nmat=PRODUCT_NMAT)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(REACTANT_NMAT, self.constraint.reactant_nmat)
        self.assertEqual(PRODUCT_NMAT, self.constraint.product_nmat)

    def testEq(self):
        if IGNORE_TEST:
            return
        constraint = DummyConstraint(reactant_nmat=REACTANT_NMAT, product_nmat=PRODUCT_NMAT)
        self.assertTrue(self.constraint == constraint)
        #
        constraint = DummyConstraint(reactant_nmat=PRODUCT_NMAT, product_nmat=REACTANT_NMAT)
        self.assertFalse(self.constraint == constraint)

    def testCopyEqual(self):
        if IGNORE_TEST:
            return
        constraint = self.constraint.copy()
        self.assertTrue(self.constraint == constraint.copy())
        #
        self.constraint.reactant_nmat.values[0,0] = 100
        self.assertFalse(self.constraint == constraint.copy())

    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        serialization_str = self.constraint.serialize()
        constraint = DummyConstraint.deserialize(serialization_str)
        self.assertEqual(self.constraint, constraint)
    
    def testClassifyReactions(self):
        if IGNORE_TEST:
            return
        reaction_classifications = self.constraint.classifyReactions()
        self.assertTrue(all([isinstance(rc, ReactionClassification) for rc in reaction_classifications]))

    def testCalculateCompatibility(self):
        if IGNORE_TEST:
            return
        result, _ = self.constraint.calculateBooleanCompatibilityVector(self.constraint.equality_nmat,
              self.constraint.equality_nmat, is_equality=True)
        num_true = np.sum(result)
        self.assertEqual(num_true, self.constraint.equality_nmat.num_row)
        #
        result, _ = self.constraint.calculateBooleanCompatibilityVector(self.constraint.numerical_inequality_nmat,
              self.constraint.equality_nmat, is_equality=True)
        num_true = np.sum(result)
        self.assertEqual(num_true, 0)
        #
        result, _ = self.constraint.calculateBooleanCompatibilityVector(self.constraint.equality_nmat,
              self.constraint.numerical_inequality_nmat, is_equality=False)
        num_true = np.sum(result)
        self.assertEqual(num_true, 5)

    def testMakeCompatibilityCollection(self):
        if IGNORE_TEST:
            return
        compatibility_collection = self.constraint.makeCompatibilityCollection(self.constraint).compatibility_collection
        self.assertTrue(np.isclose(compatibility_collection.log10_num_assignment, 0))
    
    def testMakeCompatibilityCollectionScale(self):
        if IGNORE_TEST:
            return
        scale = 10  # scaling factor
        constraint = ScalableDummyConstraint(50)
        new_constraint = constraint.scale(scale)
        compatibility_collection = constraint.makeCompatibilityCollection(new_constraint).compatibility_collection
        trues = [len(lst) >= scale
                   for lst in compatibility_collection.compatibilities]
        self.assertTrue(all(trues))

    def testCalculateLog10UnconstrainedPermutation(self):
        if IGNORE_TEST:
            return
        for size in range(3, 20):
            log10_permutation = self.constraint.calculateLog10UnconstrainedPermutation(size, size)
            self.assertTrue(np.isclose(log10_permutation, 2*np.log10(factorial(size))))

    def testExpandReductionInSize(self):
        if IGNORE_TEST:
            return
        fill_size = 2
        for size in range(3, 20):
            network = Network.makeRandomNetworkByReactionType(size, size)
            large_network = network.fill(num_fill_reaction=fill_size*size, num_fill_species=fill_size*size)
            large_constraint = ReactionConstraint(large_network.reactant_nmat, large_network.product_nmat)
            constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat)
            compatibility_collection = constraint.makeCompatibilityCollection(large_constraint).compatibility_collection
            result = compatibility_collection.expand()
            self.assertEqual(result[0].shape[1], size)


if __name__ == '__main__':
    unittest.main(failfast=True)


from pySubnetSB import constants as cn  # type: ignore
from pySubnetSB.network import Network # type: ignore
from pySubnetSB.network_collection import NetworkCollection # type: ignore
from pySubnetSB.matrix import Matrix  # type: ignore

import copy
import os
import pandas as pd # type: ignore
import numpy as np # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
COLLECTION_SIZE = 10
NETWORK_COLLECTION = NetworkCollection.makeRandomCollection(num_network=COLLECTION_SIZE)
XML_DIR = os.path.join(cn.TEST_DIR, "xml_files")


#############################
# Tests
#############################
class TestNetworkCollection(unittest.TestCase):

    def setUp(self):
        self.collection = copy.deepcopy(NETWORK_COLLECTION)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(len(self.collection) == COLLECTION_SIZE)

    def testRepr(self):
        if IGNORE_TEST:
            return
        self.assertTrue(isinstance(str(self.collection), str))

    def testMakeRandomCollection(self):
        if IGNORE_TEST:
            return
        size = 10
        collection = NetworkCollection.makeRandomCollection(num_network=size)
        self.assertTrue(len(collection) == size)

    def makeStructurallyIdenticalCollection(self, num_network:int=5, num_row:int=5, num_column:int=7):
        array1 = Matrix.makeTrinaryMatrix(num_row=num_row, num_column=num_column)
        array2 = Matrix.makeTrinaryMatrix(num_row=num_row, num_column=num_column)
        network = Network(array1, array2)
        networks = [network.permute()[0] for _ in range(num_network)]
        return NetworkCollection(networks)
    
    def testAdd(self):
        if IGNORE_TEST:
            return
        collection1 = self.makeStructurallyIdenticalCollection(num_network=15)
        collection2 = self.makeStructurallyIdenticalCollection()
        collection = collection1 + collection2
        self.assertTrue(len(collection) == len(collection1) + len(collection2))
        self.assertEqual(len(collection), str(collection).count("---") + 1)

    def testMakeFromAntimonyDirectory(self):
        if IGNORE_TEST:
            return
        directory = os.path.join(cn.TEST_DIR, "oscillators")
        network_collection = NetworkCollection.makeFromAntimonyDirectory(directory)
        self.assertTrue(len(network_collection) > 0)

    def testMakeFromAntimonyDirectoryXML(self):
        if IGNORE_TEST:
            return
        network_collection = NetworkCollection.makeFromAntimonyDirectory(XML_DIR, batch_size=-1)
        self.assertTrue(len(network_collection) > 0)

    def checkSerializeDeserialize(self, collection:NetworkCollection):
        serialization_str = collection.serialize()
        self.assertTrue(isinstance(serialization_str, str))
        #
        new_collection = NetworkCollection.deserialize(serialization_str)
        self.assertTrue(isinstance(new_collection, NetworkCollection))
        self.assertTrue(collection == new_collection)
    
    def testSerializeDeserialize1(self):
        if IGNORE_TEST:
            return
        self.checkSerializeDeserialize(self.collection)

    def testSerializeDeserialize2(self):
        if IGNORE_TEST:
            return
        def test(num_network:int=5, array_size:int=5):
            collection = NetworkCollection.makeRandomCollection(num_species=array_size, num_reaction=array_size,
                num_network=num_network)
            self.checkSerializeDeserialize(collection)
        #
        test()
        test(num_network=10, array_size=10)



if __name__ == '__main__':
    unittest.main(failfast=True)


'''Numpy 2 dimensional array with information about rows and columns.'''
from pySubnetSB.matrix import Matrix  # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import collections
import json
import pandas as pd  # type: ignore
import numpy as np
from typing import Optional, Union, List, Any


SubsetResult = collections.namedtuple('SubsetResult', ['named_matrix', 'row_idxs', 'column_idxs'])
RandomizeResult = collections.namedtuple('RandomizeResult', ['named_matrix', 'row_perm', 'column_perm'])


class NamedMatrix(Matrix):

    def __init__(self, array: np.ndarray,
                 row_names:Optional[np.ndarray]=None,
                 column_names:Optional[np.ndarray]=None,
                 row_description:str = "",
                 column_description:str = ""):
        """

        Args:
            matrix (np.ndarray): 2d numpy array
            row_names (np.ndarray): convenient identifier for rows
            column_names (np.ndarray): convenient identifier for columns
            row_description (str, optional): Name of the row. Defaults to "". Name applied to the rows.
            column_description (str, optional): Name of the column. Defaults to "". Name applied to the columns.
        """
        # Most properties are assigned on first reference since a NamedMatrix may be used only
        # as a shallow container for np.ndarray
        super().__init__(array)
        self.row_description = row_description
        self.column_description = column_description
        self._row_names:Optional[np.ndarray] = row_names
        self._column_names:Optional[np.ndarray] = column_names
        # The following are deferred execution for efficiency considerations
        self._dataframe:Optional[pd.DataFrame] = None

    @property
    def row_names(self)->np.ndarray:
        if self._row_names is None:
            self._row_names = np.array([str(n) for n in range(self.num_row)])
        else:
            if len(self._row_names) != np.shape(self.values)[0]:
                raise ValueError("Row names must have the same number of elements as the number of rows")
            self._row_names = np.array(self._row_names)
        return self._row_names  # type: ignore
    
    @property
    def column_names(self)->np.ndarray:
        if self._column_names is None:
            self._column_names = np.array([str(n) for n in range(self.num_column)])
        else:
            if len(self._column_names) != np.shape(self.values)[1]:
                raise ValueError("Column names must have the same number of elements as the number of columns")
            self._column_names = np.array(self._column_names)
        return self._column_names  # type: ignore
    
    @property
    def dataframe(self)->pd.DataFrame:
        if self._dataframe is None:
            #reduced_named_matrix = self._deleteZeroRowsColumns()
            reduced_named_matrix = self
            if len(reduced_named_matrix.values) == 0:
                return pd.DataFrame()
            self._dataframe = pd.DataFrame(reduced_named_matrix.values)
            if len(reduced_named_matrix.row_names.shape) == 1:
                self._dataframe.index = reduced_named_matrix.row_names
            if len(reduced_named_matrix.column_names.shape) == 1:
                self._dataframe.columns = reduced_named_matrix.column_names
            self._dataframe.index.name = self.row_description
            self._dataframe.columns.name = self.column_description
        return self._dataframe
    
    def copy(self)->'NamedMatrix':
        """
        Create a copy of the NamedMatrix.

        Returns:
            NamedMatrix: A copy of the NamedMatrix.
        """
        return NamedMatrix(self.values.copy(), row_names=self.row_names.copy(),
                           column_names=self.column_names.copy(),
                           row_description=self.row_description, column_description=self.column_description)
    
    def __eq__(self, other):
        """
        Compare the properties of the two NamedMatrix objects.

        Args:
            other (_type_): _description_

        Returns:
            _type_: _description_
        """
        if not super().__eq__(other):
            return False
        for attr in ['row_names', 'column_names', 'row_description', 'column_description']:
            if not np.all(getattr(self, attr) == getattr(other, attr)):
                return False
        return True

    def _deleteZeroRowsColumns(self)->'NamedMatrix':
        """
        Delete rows and columns that are all zeros.

        Returns:
            NamedMatrix: New NamedMatrix with zero rows and columns removed.
        """
        def findIndices(matrix: np.ndarray)->np.ndarray[int]:  # type: ignore
            # Finds inidices of non-zero rows
            indices = []   # Indices to delete
            for idx, array in enumerate(matrix):
                if not np.allclose(array, 0):
                    indices.append(idx)
            return np.array(indices)
        #
        row_idxs = findIndices(self.values)
        if len(row_idxs) == 0:
            return NamedMatrix(np.array([]), np.array([]), np.array([]))
        column_idxs = findIndices(self.values.T)
        matrix = self.values.copy()
        matrix = matrix[row_idxs, :]
        matrix = matrix[:, column_idxs]
        row_names = self.row_names[row_idxs]
        column_names = self.column_names[column_idxs]
        return NamedMatrix(matrix, row_names=row_names, column_names=column_names,
                           row_description=self.row_description, column_description=self.column_description)
    
    def template(self, matrix:Optional[Union[np.ndarray, Matrix]]=None)->'NamedMatrix':
        """
        Create a new NamedMatrix with the same row and column names but with a new matrix.

        Args:
            matrix (np.ndarray): New matrix to use. If None, then self is used.

        Returns:
            NamedMatrix: New NamedMatrix with the same row and column names but with a new matrix.
        """
        if matrix is None:
            matrix = self.values.copy()
        if isinstance(matrix, Matrix):
            matrix = matrix.values
        if not np.allclose(matrix.shape, self.values.shape):
            raise ValueError("Matrix shape must be the same as the original matrix")
        return NamedMatrix(matrix, row_names=self.row_names, column_names=self.column_names,
                            row_description=self.row_description, column_description=self.column_description)
    
    def isCompatible(self, other:'NamedMatrix')->bool:
        if not np.allclose(self.values.shape, other.values.shape):
            return False
        is_true =  np.all(self.row_names == other.row_names) and np.all(self.column_names == other.column_names) 
        return bool(is_true)
    
    def __repr__(self):
        return self.dataframe.__repr__()
    
    def __le__(self, other)->bool:
        if not self.isCompatible(other):
            return False
        return bool(np.all(self.values <= other.values))
        
    def getSubNamedMatrix(self, row_names:Optional[Union[np.ndarray, list]]=None,
                     column_names:Optional[Union[np.ndarray, list]]=None)->SubsetResult:
        """
        Create an ndarray that is a subset of the rows in the NamedMatrix.

        Args:
            row_names (list): List of row names to keep. If None, keep all.
            column_names (list): List of row names to keep. If None, keep all.

        Returns:
            SubsetResult (readonly values)
        """
        def cleanName(name):
            if name[0] in ["[", "("]:
                new_name = name[1:]
            else:
                new_name = name
            if name[-1] in ["]", ")"]:
                new_name = new_name[:-1]
            return new_name.replace(",", "")
        def findIndices(sub_names,
                        all_names=None)->np.ndarray:
            if all_names is None:
                return np.array(range(len(sub_names)))
            sub_names_lst = [cleanName(str(n)) for n in sub_names]
            all_names_lst = [cleanName(str(n)) for n in all_names]
            indices = np.repeat(-1, len(sub_names))
            # Find the indices of the names in the other_names and place them in the correct order
            for sub_idx, sub_name in enumerate(sub_names_lst):
                if any([np.all(sub_name == o) for o in all_names_lst]):
                    all_names_idx = all_names_lst.index(sub_name)
                    indices[sub_idx] = all_names_idx
                else:
                    raise ValueError(f'Could not find name {sub_name} in other names!')
            return np.array(indices)
        #
        if row_names is None:
            row_idxs = np.array(range(self.num_row))
        else:
            row_idxs = findIndices(row_names, self.row_names)
        if column_names is None:
            column_idxs = np.array(range(self.num_column))
        else:
            column_idxs = findIndices(column_names, self.column_names)
        new_values = self.values[row_idxs, :].copy()
        new_values = new_values[:, column_idxs]
        named_matrix = NamedMatrix(new_values, row_names=self.row_names[row_idxs],
                                   column_names=self.column_names[column_idxs])
        return SubsetResult(named_matrix=named_matrix,
                            row_idxs=row_idxs, column_idxs=column_idxs)
    
    def getRowIdx(self, row_name:str)->int:
        """
        Obtain the index of the row by name.

        Args:
            row_name (str): _description_

        Returns:
            int: _description_
        """
        return list(self.row_names).index(row_name)
    
    def getColumnIdx(self, column_name:str)->int:
        """
        Obtain the index of the column by name.

        Args:
            column_name (str): _description_

        Returns:
            int: _description_
        """
        return list(self.column_names).index(column_name)
    
    def getValueByNames(self, row_name:str, column_name:str)->Any:
        """
        Obtains the value of the NamedMatrix by row and column names.

        Args:
            row_name (str): _description_
            column_name (str): _description_

        Returns:
            Any: _description_
        """
        return self.values[self.getRowIdx(row_name), self.getColumnIdx(column_name)]
        
    def getSubMatrix(self, row_idxs:np.ndarray, column_idxs:np.ndarray)->Matrix:
        """
        Create an ndarray that is a subset of the rows in the NamedMatrix.

        Args:
            row_idxs (ndarray): row indices to keep
            column_idxs (ndarray): column indices to keep

        Returns:
            Matrix
        """
        return Matrix(self.values[row_idxs, :][:, column_idxs])

    def transpose(self):
        return NamedMatrix(self.values.T, row_names=self.column_names,
                           column_names=self.row_names,
                           row_description=self.column_description, column_description=self.row_description)
    
    def serialize(self)->str:
        """Serializes the NamedMatrix."""
        return json.dumps({cn.S_ID: str(self.__class__),
                           cn.S_ROW_NAMES: self.row_names.tolist(),
                           cn.S_COLUMN_NAMES: self.column_names.tolist(),
                           cn.S_ROW_DESCRIPTION: self.row_description,
                           cn.S_COLUMN_DESCRIPTION: self.column_description,
                           cn.S_VALUES: self.values.tolist()})
    
    @classmethod
    def deserialize(cls, string:str)->'NamedMatrix':
        """
        Deserializes the NamedMatrix.

        Args:
            string (str): A JSON string.

        Returns:
            NamedMatrix: A NamedMatrix object.
        """
        dct = json.loads(string)
        if not str(cls) in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls} but got {dct[cn.S_ID]}")
        return cls(np.array(dct[cn.S_VALUES]), row_names=np.array(dct[cn.S_ROW_NAMES]),
                   column_names=np.array(dct[cn.S_COLUMN_NAMES]),
                   row_description=dct[cn.S_ROW_DESCRIPTION],
                   column_description=dct[cn.S_COLUMN_DESCRIPTION])
    
    @classmethod
    def hstack(cls, named_matrices:List['NamedMatrix'], is_rename:bool=False)->'NamedMatrix':
        """
        Stacks horizontally a list of of compatible NamedMatrix objects.

        Args:
            named_matrices (List[NamedMatrix])
            is_rename: bool (rename the columns)

        Returns:
            NamedMatrix

        Raises:
            ValueError: If the row names are not the same. 
        """
        # Check compatibility
        first_num_row = named_matrices[0].num_row
        first_row_names = named_matrices[0].row_names
        first_row_description = named_matrices[0].row_description
        first_column_description = named_matrices[0].column_description
        for named_matrix in named_matrices[1:]:
            if first_column_description != named_matrix.column_description:
                if len(first_column_description) > 0 and len(named_matrix.column_description) > 0:
                    raise ValueError("Column descriptions must be the same!")
            if first_row_description != named_matrix.row_description:
                if len(first_row_description) > 0 and len(named_matrix.row_description) > 0:
                    raise ValueError("Row descriptions must be the same!")
            if not np.all(first_row_names == named_matrix.row_names):
                raise ValueError("Row descriptions must be the same!")
            if first_num_row != named_matrix.num_row:
                raise ValueError("Number of rows must be the same!")
        # Stack the values and construct the NamedMatrix
        values = np.hstack([named_matrix.values for named_matrix in named_matrices])
        if is_rename:
            column_names = None
        column_names = np.concatenate([named_matrix.column_names for named_matrix in named_matrices])
        return NamedMatrix(values, row_names=first_row_names, column_names=column_names,
                           row_description=first_row_description,
                           column_description=first_column_description)
    
    @classmethod
    def vstack(cls, named_matrices:List['NamedMatrix'], is_rename:bool=False)->'NamedMatrix':
        """
        Stacks vertically a list of of compatible NamedMatrix objects.

        Args:
            named_matrices (List[NamedMatrix])

        Returns:
            NamedMatrix

        Raises:
            ValueError: If the row names are not the same. 
        """
        # Check compatibility
        first_num_column = named_matrices[0].num_column
        first_column_names = named_matrices[0].column_names
        first_column_description = named_matrices[0].column_description
        first_row_description = named_matrices[0].column_description
        for named_matrix in named_matrices[1:]:
            if first_row_description != named_matrix.row_description:
                raise ValueError("Row descriptions must be the same!")
            if first_column_description != named_matrix.column_description:
                raise ValueError("Column descriptions must be the same!")
            if not np.all(first_column_names == named_matrix.column_names):
                raise ValueError("Column descriptions must be the same!")
            if first_num_column != named_matrix.num_column:
                raise ValueError("Number of columns must be the same!")
        # Stack the values and construct the NamedMatrix
        values = np.vstack([named_matrix.values for named_matrix in named_matrices])
        if is_rename:
            row_names = None
        else:
            row_names = np.concatenate([named_matrix.row_names for named_matrix in named_matrices])
        return NamedMatrix(values, column_names=first_column_names, row_names=row_names,
                           column_description=first_column_description,
                           row_description=first_row_description)
    
    @classmethod
    def makeRandom(cls, num_row:int, num_column:int)->'NamedMatrix':
        """
        Create a random NamedMatrix.

        Args:
            num_row (int): Number of rows
            num_column (int): Number of columns

        Returns:
            NamedMatrix: A random NamedMatrix.
        """
        return NamedMatrix(np.random.rand(num_row, num_column))
    
    def vmerge(self, other:'NamedMatrix')->'NamedMatrix':
        """
        Append other to current, padding if necessary. Must have distinct column names and the
        same row_description and column_description.
        
        Args:
            other (Network): Network to merge to the right of current network
            is_pad_bottom (bool, optional): If True, pad the bottom of to make networks have the same number of rows.

        Returns:
            NamedMatrix
        """
        def extend(named_matrix:'NamedMatrix', name_arr:List[str])->'NamedMatrix':
            # Add missing rows; sort in order by row names
            row_names = list(name_arr)
            missing_row_names = list(set(row_names) - set(named_matrix.row_names))
            if len(missing_row_names) > 0:
                missing_rows = np.zeros((len(missing_row_names), named_matrix.num_column))
            else:
                missing_rows = np.array([])
            arr = named_matrix.values.copy()
            if len(missing_rows) > 0:
                arr = np.vstack([arr, missing_rows])
            current_row_names = list(named_matrix.row_names) + missing_row_names
            permutation = [current_row_names.index(n) for n in row_names]
            permuted_arr = arr[permutation, :]
            return NamedMatrix(permuted_arr, row_names=np.array(row_names), column_names=named_matrix.column_names,
                  row_description=named_matrix.row_description, column_description=named_matrix.column_description)
        #
        if self.row_description != other.row_description:
            raise ValueError("Row descriptions must be the same!")
        if len(set(self.column_names).intersection(set(other.column_names))) > 0:
            raise ValueError("Column names must be unique!")
        # Find the set of row names
        row_names = np.unique(np.concatenate([self.row_names, other.row_names]))
        # 
        self_nm = extend(self, row_names)
        other_nm = extend(other, row_names)
        return self_nm.hstack([self_nm, other_nm])
    
    def randomize(self, row_perm:Optional[np.ndarray]=None, column_perm:Optional[np.ndarray]=None)->RandomizeResult:
        """
        Randomize the order of rows and columns of the NamedMatrix.

        Returns:
            RandomizeResult: NamedMatrix, row permutation, column permutation
        """
        if row_perm is None:
            row_perm = np.random.permutation(self.num_row)
        if column_perm is None:
            column_perm = np.random.permutation(self.num_column)
        arr = self.values.copy()
        arr = arr[row_perm, :]
        arr = arr[:, column_perm]
        nmat = NamedMatrix(arr, self.row_names[row_perm], self.column_names[column_perm],
                           row_description=self.row_description, column_description=self.column_description)
        return RandomizeResult(named_matrix=nmat, row_perm=row_perm, column_perm=column_perm)
    
    def sort(self)->'NamedMatrix':
        """
        Sorts the order of rows and columns of the NamedMatrix by row and column names.

        Returns:
            NamedMatrix: A NamedMatrix with random values.
        """
        row_perm = np.argsort(self.row_names)
        column_perm = np.argsort(self.column_names)
        arr = self.values.copy()
        arr = arr[row_perm, :]
        arr = arr[:, column_perm]
        return NamedMatrix(arr, self.row_names[row_perm], self.column_names[column_perm],
                           row_description=self.row_description, column_description=self.column_description)


NULL_NMAT = NamedMatrix(np.array([[]]))


from pySubnetSB.identity_hash_benchmark import IdentityHashBenchmark, COLUMNS  # type: ignore

import pandas as pd # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
REACTION_NUMS  = [2, 5, 10, 15]
SPECIES_NUMS = [2, 5, 10, 15]
NUM_NETWORK = 100


#############################
# Tests
#############################
class TestBenchmark(unittest.TestCase):

    def setUp(self):
        self.benchmark = IdentityHashBenchmark(SPECIES_NUMS, REACTION_NUMS, num_network=NUM_NETWORK)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.benchmark.num_network, NUM_NETWORK)

    def testCalculateHashStatistics(self):
        if IGNORE_TEST:
            return
        df = self.benchmark.calculateHashStatistics()
        trues = [c in df.columns for c in COLUMNS]
        self.assertTrue(all(trues))
        self.assertEqual(len(df), len(SPECIES_NUMS) * len(REACTION_NUMS))

    def testPlotHashStatistics(self):
        if IGNORE_TEST:
            return
        _ = self.benchmark.plotHashStatistics(font_size=14, is_plot=IS_PLOT)

    def testPlotHashStatisticsScale(self):
        if IGNORE_TEST:
            return
        species_nums = range(2, 22, 2)
        reaction_nums = range(2, 22, 2)
        benchmark = IdentityHashBenchmark(species_nums, reaction_nums, num_network=1000)
        _ = benchmark.plotHashStatistics(font_size=14, is_plot=IS_PLOT)



if __name__ == '__main__':
    unittest.main(failfast=True)


'''Constraints for reactions.'''

"""
Enumerated constraints.
    Reaction type (equality)
    Number precdecessors of each type
    Number of successors of each type
"""

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB.constraint import Constraint, NULL_NMAT # type: ignore
from pySubnetSB.constraint_option_collection import ReactionConstraintOptionCollection # type: ignore

import numpy as np
from typing import List


#####################################
class ReactionConstraint(Constraint):

    def __init__(self, reactant_nmat:NamedMatrix, product_nmat:NamedMatrix,
                 reaction_constraint_option_collection:ReactionConstraintOptionCollection=ReactionConstraintOptionCollection(),
                 is_subnet:bool=True):
        """
        Args:
            reactant_nmat (NamedMatrix)
            product_nmat (NamedMatrix)
        """
        super().__init__(reactant_nmat=reactant_nmat, product_nmat=product_nmat, is_subnet=is_subnet)
        self._is_initialized = False
        self._numerical_enumerated_nmat = NULL_NMAT  
        self._numerical_categorical_nmat = NULL_NMAT
        self._bitwise_enumerated_nmat = NULL_NMAT
        self._bitwise_categorical_nmat = NULL_NMAT
        self._one_step_nmat = NULL_NMAT
        self._to_reaction_arr = np.eye(self.num_reaction)
        self.option_collection = reaction_constraint_option_collection

    def _initialize(self):
        """
        Creates the constraint matrices based on the constraint options.
        """
        if self._is_initialized:
            return
        numerical_enumerated_nmats = []
        if self.option_collection.isTrue("is_make_successor_predecessor_constraint_matrix"):
            numerical_enumerated_nmats.append(self.makeSuccessorPredecessorConstraintMatrix())
        if self.option_collection.isTrue("is_make_n_step_constraint_matrix"):
            numerical_enumerated_nmats.append(self.makeNStepConstraintMatrix(num_step=2))
        #
        numerical_categorical_nmats = []
        if self.option_collection.isTrue("is_make_classification_constraint_matrix"):
            numerical_categorical_nmats.append(self._makeClassificationConstraintMatrix())
        if self.option_collection.isTrue("is_make_autocatalysis_constraint_matrix"):
            numerical_categorical_nmats.append(self._makeAutocatalysisConstraintMatrix())
        #
        if len(numerical_enumerated_nmats) > 0:
            self._numerical_enumerated_nmat = NamedMatrix.hstack(numerical_enumerated_nmats)
        if len(numerical_categorical_nmats) > 0:
            self._numerical_categorical_nmat = NamedMatrix.hstack(numerical_categorical_nmats)
        self._is_initialized = True

    ################# OVERLOADED PARENT CLASS METHODS #################
    @property
    def row_names(self)->List[str]:
        return self.reactant_nmat.column_names
    
    @property
    def description(self)->str:
        return "reactions"

    @property
    def numerical_enumerated_nmat(self)->NamedMatrix:
        self._initialize()
        return self._numerical_enumerated_nmat
    
    @property
    def categorical_nmat(self)->NamedMatrix:
        self._initialize()
        return self._numerical_categorical_nmat

    @property
    def bitwise_enumerated_nmat(self)->NamedMatrix:
        return self._bitwise_enumerated_nmat
    
    @property
    def one_step_nmat(self)->NamedMatrix:
        """Calculates the successor matrix for the species monopartite graph.

        Returns:
            np.ndarray: _description_
        """
        # Create the monopartite graph
        if self._one_step_nmat is NULL_NMAT:
            incoming_arr = self.reactant_nmat.values.T
            outgoing_arr = self.product_nmat.values.T
            arr = np.sign(np.matmul(outgoing_arr, incoming_arr.T)).astype(int)
            self._one_step_nmat = NamedMatrix(arr, row_names=self.reactant_nmat.column_names,
                                row_description='reactions',
                                column_description='reactions',
                                column_names=self.reactant_nmat.column_names)
        return self._one_step_nmat
    
    @property
    def to_reaction_arr(self)->np.ndarray:
        return self._to_reaction_arr
    
    ##################################

    def __repr__(self)->str:
        return "Reaction--" + super().__repr__()
    
    def _makeClassificationConstraintMatrix(self)->NamedMatrix:
        """Make constraints for the reaction type. These are the number of
        reactions of each type.

        Returns:
            NamedMatrix: Rows are reactions, columns are constraints by count of reaction type.
        """
        array = np.array([self.reaction_classification_arr[n].encoding for n in range(self.num_reaction)])
        array = np.reshape(array, (len(array), 1)).astype(int)
        column_names = ["reaction_classifications"]
        named_matrix = NamedMatrix(array, row_names=self.reactant_nmat.column_names,
                           row_description=self.reactant_nmat.column_description,
                           column_description="constraints",
                           column_names=column_names)
        return named_matrix
    
    def _makeAutocatalysisConstraintMatrix(self)->NamedMatrix:
        """Make constraints for autocatalysis. These are the number of
        reactions of each type that are autocatalytic.

        Returns:
            NamedMatrix: Rows are reactions, columns are constraints by count of reaction type.
        """
        autocatalysis_arr = np.sum(self.reactant_nmat.values * self.product_nmat.values, axis=0)
        autocatalysis_arr = np.sign(autocatalysis_arr)
        autocatalysis_arr = np.reshape(autocatalysis_arr, (len(autocatalysis_arr), 1)).astype(int)
        column_names = ["autocatalysis"]
        named_matrix = NamedMatrix(autocatalysis_arr, row_names=self.reactant_nmat.column_names,
                           row_description=self.reactant_nmat.column_description,
                           column_description="constraints",
                           column_names=column_names)
        return named_matrix


import pySubnetSB.api as api # type: ignore
from pySubnetSB.network import Network  # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import os
import numpy as np
import tellurium as te # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
MODEL = """
J1: A -> B; k1*A

k1 = 0.1
A = 10
B = 0
"""
MODEL_RR = te.loada(MODEL)
SBML_PATH = os.path.join(cn.TEST_DIR, "test_api.sbml")
ANT_PATH = os.path.join(cn.TEST_DIR, "test_api.ant")
SERIALIZATION_PATH = os.path.join(cn.TEST_DIR, "test_api.txt")
REMOVE_FILES = [SBML_PATH, ANT_PATH, SERIALIZATION_PATH]
MODEL_DIR = os.path.join(cn.TEST_DIR, "oscillators")
URL = "https://www.ebi.ac.uk/biomodels/services/download/get-files/MODEL1701090001/3/BIOMD0000000695_url.xml"

#############################
# Tests
#############################

class TestModelSpecification(unittest.TestCase):

    def setUp(self):
        self.remove()
        self.specification = api.ModelSpecification(MODEL, specification_type="antstr")

    def tearDown(self):
        self.remove()

    def remove(self):
        for path in REMOVE_FILES:
            if os.path.isfile(path):
                os.remove(path)

    def testGetNetwork(self):
        if IGNORE_TEST:
            return
        def test(model, specification_type):
            self.specification = api.ModelSpecification(model, specification_type=specification_type)
            network = self.specification.getNetwork()
            self.assertTrue(isinstance(network, Network))
        #
        test(URL, "sbmlurl")
        #
        path = os.path.join(cn.TEST_DIR, "xml_files/BIOMD0000000033.xml")
        test(path, "sbmlfile")
        #
        rr = te.loadSBMLModel(URL)
        test(rr, "roadrunner")
        #
        rr = te.loadSBMLModel(URL)
        test(rr.getSBML(), "sbmlstr")
        #
        path = os.path.join(cn.TEST_DIR, "oscillators/bestmodel_jJykOfGq0Kgy")
        test(path, "antfile")
        #
        with open(path, "r") as fd:
            lines = fd.readlines()
        model = "\n".join(lines)
        test(model, "antstr")

    def testMakeNetworkAntimonyStr(self):
        if IGNORE_TEST:
            return
        network = self.specification.makeNetwork(MODEL)
        self.assertTrue(np.all(network.species_names == ["A", "B"]))
        self.assertTrue(network.reaction_names == ["J1"])
    
    def testMakeNetworkAntimonyFile(self):
        if IGNORE_TEST:
            return
        with open(ANT_PATH, "w") as fd:
            fd.write(MODEL)
        network = self.specification.makeNetwork(ANT_PATH, specification_type="antfile")
        self.assertTrue(np.all(network.species_names == ["A", "B"]))
        self.assertTrue(network.reaction_names == ["J1"])

    def testMakeNetworkSBMLStr(self):
        if IGNORE_TEST:
            return
        sbml_str = MODEL_RR.getSBML()
        network = api.ModelSpecification.makeNetwork(sbml_str, specification_type="sbmlstr")
        self.assertTrue(np.all(network.species_names == ["A", "B"]))
        self.assertTrue(network.reaction_names == ["J1"])

    def testMakeNetworkSBMLFile(self):
        if IGNORE_TEST:
            return
        sbml_str = MODEL_RR.getSBML()
        with open(SBML_PATH, "w") as fd:
            fd.write(sbml_str)
        network = self.specification.makeNetwork(SBML_PATH, specification_type="sbmlfile")
        self.assertTrue(np.all(network.species_names == ["A", "B"]))
        self.assertTrue(network.reaction_names == ["J1"])
        

class TestFunctions(unittest.TestCase):

    @staticmethod
    def optionIter(excludes=None):
        if excludes is None:
            excludes = []
        dct = {}
        dcts = []
        for is_subnet in [True, False]:
            if not 'is_subnet' in excludes:
                dct['is_subnet'] = is_subnet
            for num_process in [1, -1]:
                if not 'num_process' in excludes:
                    dct['num_process'] = num_process
                for max_num_assignment in [1e10, 1]:
                    if not 'max_num_mapping_pair' in excludes:
                        dct['max_num_mapping_pair'] = max_num_assignment
                    for identity in [cn.ID_STRONG, cn.ID_WEAK]:
                        if not 'identity' in excludes:
                            dct['identity'] = identity
                        is_duplicate = False
                        for dct1 in dcts:
                            if dct1 == dct:
                                is_duplicate = True
                                break
                        if not is_duplicate:
                            yield dct

    def testOptionIter(self):
        if IGNORE_TEST:
            return
        for dct in self.optionIter(["is_subnet", "num_process"]):
            self.assertFalse("num_process" in dct)

    def testFindReferenceInTarget(self):
        if IGNORE_TEST:
            return
        iter = self.optionIter()
        for dct in iter:
            result = api.findReferenceInTarget(MODEL, MODEL, **dct)
            self.assertTrue(len(result.assignment_pairs) == 1)
            self.assertTrue(np.all(result.assignment_pairs[0].species_assignment == [0, 1]))
            self.assertTrue(np.all(result.assignment_pairs[0].reaction_assignment == [0]))
            self.assertTrue(result.is_truncated == False)

    def testClusterStructurallyIdenticalModelsInDirectory(self):
        if IGNORE_TEST:
            return
        ffiles = [f for f in os.listdir(MODEL_DIR) if "best" in f]
        iter = self.optionIter(["is_subnet", "num_process"])
        for dct in iter:
            if "max_num_mapping_pair" in dct:
                dct["max_num_assignment"] = dct["max_num_mapping_pair"]
                dct.pop("max_num_mapping_pair")
            df = api.clusterStructurallyIdenticalModelsInDirectory(MODEL_DIR, **dct)
            self.assertEqual(len(df), len(ffiles))

    def testFindReferencesInTargets(self):
        if IGNORE_TEST:
            return
        count = len([f for f in os.listdir(MODEL_DIR) if "best" in f])
        iter = self.optionIter(["is_subnet"])
        for dct in iter:
            df = api.findReferencesInTargets(MODEL_DIR, MODEL_DIR, **dct)
            self.assertEqual(len(df), count**2)
            num_match = np.sum([len(v) > 0 for v in df[cn.FINDER_INDUCED_NETWORK]]) \
                + np.sum([v for v in df[cn.FINDER_IS_TRUNCATED]])
            self.assertGreaterEqual(num_match, count)  # May be networks that are a subnet of another

    def testMakeSerializationFile(self):
        if IGNORE_TEST:
            return
        api.makeSerializationFile(MODEL_DIR, SERIALIZATION_PATH, is_report=IS_PLOT)
        self.assertTrue(os.path.isfile(SERIALIZATION_PATH))

    def testGetNetworkCollection(self):
        if IGNORE_TEST:
            return
        network_collection = api._getNetworkCollection(MODEL_DIR)
        self.assertTrue(len(network_collection) > 0)
        #
        url = "http://raw.githubusercontent.com/ModelEngineering/pySubnetSB/main/examples/target_serialized.txt"
        network_collection = api._getNetworkCollection(url)
        self.assertTrue(len(network_collection) > 0)


if __name__ == '__main__':
    unittest.main()


from pySubnetSB.constraint import Constraint, ReactionClassification, CompatibilityCollection    # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB.compatibility_collection_inspector import CompatibilityCollectionInspector  # type: ignore

import numpy as np
import unittest


IGNORE_TEST = False
IS_PLOT = False
NUM_SPECIES, NUM_REACTION = 3, 3
SPECIES_NAMES = ["S" + str(i) for i in range(NUM_SPECIES)]
REACTION_NAMES = ["J" + str(i) for i in range(NUM_REACTION)]
REFERENCE_NETWORK = Network.makeRandomNetworkByReactionType(NUM_SPECIES, NUM_REACTION)
TARGET_NETWORK = REFERENCE_NETWORK.fill(num_fill_reaction=1, num_fill_species=1)


#############################
class TestCompatibilityCollectionInspector(unittest.TestCase):

    def setUp(self):
        self.inspector = CompatibilityCollectionInspector(REFERENCE_NETWORK, TARGET_NETWORK, is_species=True)

    def testMakeDiagnosticLabelMatrix(self):
        if IGNORE_TEST:
            return
        label_nmat = self.inspector._makeDiagnosticLabelMatrix()
        self.assertEqual(len(label_nmat), REFERENCE_NETWORK.num_species * TARGET_NETWORK.num_species)
        self.assertEqual(label_nmat.num_column, 2)
    
    def testGetTrueRowsInDiagnosticMatrix(self):
        if IGNORE_TEST:
            return
        fill_size = 3
        for is_species in [True, False]:
            for size in range(4, 10):
                reference_network = Network.makeRandomNetworkByReactionType(size, size)
                target_network = reference_network.fill(num_fill_reaction=fill_size, num_fill_species=fill_size)
                inspector = CompatibilityCollectionInspector(reference_network, target_network, is_species=is_species)
                true_idxs = inspector.getTrueRowsInDiagnosticMatrix()
                pair_arr = inspector.diagnostic_label_nmat.values[true_idxs, :]
                num_match = np.sum([p[0] == p[1] for p in pair_arr])
                self.assertGreaterEqual(num_match, reference_network.num_species)

    def testCalculateDiagnosticMatrix(self):
        if IGNORE_TEST:
            return
        diag_nmat = self.inspector._makeDiagnosticMatrix()
        self.assertEqual(diag_nmat.num_row, REFERENCE_NETWORK.num_species * TARGET_NETWORK.num_species)

    def testMakeColumnNameDct(self):
        if IGNORE_TEST:
            return
        column_name_dct = self.inspector._makeColumnNameDct()
        keys = list(column_name_dct.keys())
        value = column_name_dct[keys[0]]
        self.assertTrue(isinstance(column_name_dct, dict))
        self.assertTrue(isinstance(value.column_name, str))
        self.assertTrue(isinstance(value.named_matrix, NamedMatrix))
        self.assertTrue(isinstance(value.constraint_type, str))

    def testGetConstraintValue(self):
        if IGNORE_TEST:
            return
        element_name = "S0"
        for key, descriptor in self.inspector.column_name_dct.items():
            column_name = key[1]
            is_reference = key[0] == "reference"
            nmat = descriptor.named_matrix
            row_idx = list(nmat.row_names).index(element_name)
            column_idx = list(nmat.column_names).index(column_name)
            value = self.inspector._getConstraintValue(element_name, column_name, is_reference)
            self.assertTrue(np.all(nmat.values[row_idx, column_idx] == value))


if __name__ == '__main__':
    unittest.main(failfast=True)


'''Provides the API for the analysis of structural identity.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.cluster_builder import ClusterBuilder  # type: ignore
from pySubnetSB.model_serializer import ModelSerializer  # type: ignore
from pySubnetSB.network import Network, StructuralAnalysisResult  # type: ignore
from pySubnetSB.network_collection import NetworkCollection  # type: ignore
from pySubnetSB.subnet_finder import SubnetFinder  # type: ignore
import pySubnetSB.util as util # type: ignore

import os
import pandas as pd  # type: ignore
import tellurium as te  # type: ignore
import tempfile # type: ignore
from typing import Union, Tuple, Optional

# Ways in which a model can be specified
SPECIFICATION_TYPES = ["antstr", "antfile", "sbmlstr", "sbmlfile", "sbmlurl", "roadrunner"]
DEFAULT_SPECIFICATION_TYPE = "antstr"


#############################
def _getNetworkCollection(directory:str, is_report:bool=True)->Tuple[NetworkCollection, Optional[str]]:
    """Returns the NetworkCollection from the directory.

    Args:
        directory (str): Path to a directory or a serialization file
        is_report (bool): If True, report progress

    Returns:
        NetworkCollection: contains a list of networks
        Optional[str]: Path to the serialization file
    """
    if directory.endswith(".txt"):
        if "http" in directory:
            strings = util.getStringsFromURL(directory)
            networks = ModelSerializer.deserializeFromStrings(strings)
            network_collection = NetworkCollection(networks)
            return network_collection, None
        else:
            serialization_path = directory
        serializer = ModelSerializer(None, serialization_path)
        filename = None
    else:
        # Create the serialization file
        fp = tempfile.NamedTemporaryFile(delete=False)
        filename = fp.name
        fp.close()
        serializer = ModelSerializer(directory, filename)
        if is_report:
            report_interval = 10
        else:
            report_interval = None
        serializer.serialize(report_interval=report_interval)
    return serializer.deserialize(), filename


#############################
class ModelSpecification(object):
    """Specification of a model for a structural analysis."""
    def __init__(self, model:str, specification_type=DEFAULT_SPECIFICATION_TYPE)->None:
        """
        Args:
            model (str): Antimony model or SBML model or path to model
            specification_type:
                "antstr": Antimony model string
                "antfile": Antimony model file
                "sbmlstr": SBML model string
                "sbmlfile": SBML model file
                "sbmlurl": SBML model URL
                "roadrunner": RoadRunner model
        """
        self.model = model
        self.specification_type = specification_type
        if not specification_type in SPECIFICATION_TYPES:
            raise ValueError(f"Invalid model type: {specification_type}. Must be one of {SPECIFICATION_TYPES}")
        
    def __repr__(self)->str:
        return f"ModelReference({self.model}, {self.specification_type})"

    @classmethod 
    def makeNetwork(cls, model:Union[str, 'ModelSpecification'],
          specification_type:Optional[str]=DEFAULT_SPECIFICATION_TYPE)->Network:
        """Returns the Antimony model."""
        if isinstance(model, ModelSpecification):
            specification = model
        else:
            specification = cls(model, specification_type=specification_type)
        network = specification.getNetwork()
        return network
    
    def getNetwork(self)->Network:
        """Returns the Antimony model."""
        if self.specification_type not in SPECIFICATION_TYPES:
            raise ValueError(f"Invalid model type: {self.specification_type}. Must be one of {SPECIFICATION_TYPES}")
        # Check for roadrunner
        if self.specification_type == "roadrunner":
            network = Network.makeFromAntimonyStr(None, roadrunner=self.model)
            return network
        # Check for SBML
        if self.specification_type in ["sbmlstr", "sbmlfile", "sbmlurl"]:
            network = Network.makeFromSBMLFile(self.model)
            return network
        # Handle Antimony
        if self.specification_type == "antstr":
            antimony_str = self.model
        elif self.specification_type == "antfile":
            with open(self.model, "r") as fd:
                antimony_str = fd.read()
        else:
            raise ValueError(f"Invalid model type: {self.specification_type}. Must be one of {SPECIFICATION_TYPES}")
        #
        network = Network.makeFromAntimonyStr(antimony_str)
        return network


#############################
def findReferenceInTarget(
      reference_model:Union[str, ModelSpecification],
      target_model:Union[str, ModelSpecification],
      is_subnet:bool=True,
      num_process:int=-1,
      max_num_mapping_pair:int=int(1e12),
      identity:str=cn.ID_STRONG,
      is_report:bool=True)->StructuralAnalysisResult:
    """
    Searches the target model to determine if the reference model is a subnet (if is_subnet is True) or
    if the reference model is structurally identical to the target model (if is_subnet is False).

    Args:
        reference_model (str/ModelSpecification): Antimony str or other model reference
           Use for ModelSpecification:
               ModelSpecification(model, specification_type)
               specification_type:
                   "antstr": Antimony model string
                   "antfile": Antimony model file
                   "sbmlstr": SBML model string
                   "sbmlfile": SBML model file
                   "sbmlurl": SBML model URL
                   "roadrunner": RoadRunner model
        target_model (str/ModelSpecification): Antimony str or other model reference
        is_subnet (bool): If True, the reference model is a subnet of the target model;
                          otherwise, the reference model is structurally identical to the target model.
        num_process (int): Number of processes to use. If -1, use all available processors.
        max_num_mapping_pair (int): Maximum number of mapping pairs. (Internally, this is the number of assignment pairs)
        identity (str): cn.ID_STRONG (renaming) or cn.ID_WEAK (behavior equivalence)
        is_report (bool): If True, report progress

    Returns:
        StructuralAnalysisResult: List of StructuralAnalysisResult
            assignment_pairs: Assignments of target species/reactions to the reference that result in structural identity
            is_truncated: If True, the search was truncated
            num_species_candidate: Number of assignments of target species to reference species that satisfy constraints
            num_reaction_candidate: Number of assignments of target reaction to reference reaction that satisfy constraints
    """
    reference_network = ModelSpecification.makeNetwork(reference_model)
    target_network = ModelSpecification.makeNetwork(target_model)
    return reference_network.isStructurallyIdentical(
            target_network,
            is_subnet=is_subnet,
            num_process=num_process,
            identity=identity,
            max_num_assignment=max_num_mapping_pair,
            is_report=is_report)

#############################
def clusterStructurallyIdenticalModelsInDirectory(
      model_dir:str,
      identity:str=cn.ID_STRONG,
      max_num_assignment:int=int(1e12),
      is_report:bool=True)->pd.DataFrame:
    """
    Forms clusters of structurally identical models.

    Args:
        model_dir (ModelSpecification/str): Directory with Antimony models or SBML models or a serialization file
        max_num_assignment (int): Maximum number of assignment pairs
        identity (str): cn.ID_STRONG (renaming) or cn.ID_WEAK (behavior equivalence)
        is_report (bool): If True, report progress

    Returns: pd.DataFrame: DataFrame with the subnets. Columns:
        network_name: Name of first network in the cluster collection
        processing_time: Time to process this collection
        is_indeterminate: Indeterminante because too many possible assignments
        assignment_collection: List of networks in the collection
    """
    network_collection, _ = _getNetworkCollection(model_dir, is_report=is_report)
    builder = ClusterBuilder(network_collection,
               identity=identity,
               max_num_assignment=max_num_assignment,
               is_report=is_report)
    builder.cluster()
    # Construct the result
    dct:dict = {k: [] for k in [cn.S_PROCESSING_TIME, cn.S_HASH_VAL, cn.S_PROCESSED_NETWORKS, cn.S_IS_INDETERMINATE]}
    for processed_network_collection in builder.processed_network_collections:
        dct[cn.S_PROCESSING_TIME].append(processed_network_collection.processing_time)
        dct[cn.S_HASH_VAL].append(processed_network_collection.hash_val)
        network_names = [c.network_name for c in processed_network_collection.processed_networks]
        dct[cn.S_PROCESSED_NETWORKS].append(network_names)
        is_indeterminates = [c.is_indeterminate for c in processed_network_collection.processed_networks]
        dct[cn.S_IS_INDETERMINATE].append(any(is_indeterminates))
    return pd.DataFrame(dct)


#############################
def findReferencesInTargets(
      reference_dir:str,
      target_dir:str,
      identity:str=cn.ID_STRONG,
      num_process:int=-1,
      max_num_mapping_pair:int=int(1e12),
      is_report:bool=True)->pd.DataFrame:
    """
    Searches the target directory to determine if the reference directory contains subnets (if is_subnet is True) or
    if the reference directory contains models that are structurally identical to the target directory
    (if is_subnet is False).

    Args:
        reference_dir (str): Directory with Antimony models or SBML models or a serialization file
        target_dir (str): Directory with Antimony models or SBML models or a serialization file
        is_subnet (bool): If True, the reference directory contains subnets of the target directory;
                          otherwise, the reference directory contains models that are structurally identical to the target directory.
        num_process (int): Number of processes to use. If -1, use all available processors.
        max_num_mapping_pair (int): Maximum number of mapping pairs. (Internally, this is the number of assignment pairs)
        identity (str): cn.ID_STRONG (renaming) or cn.ID_WEAK (behavior equivalence)
        is_report (bool): If True, report progress

    Returns:
        pd.DataFrame: DataFrame with the subnets
                reference_name (str): Reference model name
                target_name (str): Target model name
                reference_network (str): string representation of the reference network
                induced_network (str): string representation of the induced network in the target
                name_dct (dict): Dictionary of mapping of target names to reference names for species and reactions
                                 as a JSON string.
                num_assignment_pair (int): Number of assignment pairs returned
                                            for matching target to subnet
                is_trucated (bool): True if the search is truncated
    """
    # Process the request
    reference_tempfile, target_tempfile = None, None
    try: 
        reference_collection, reference_tempfile = _getNetworkCollection(reference_dir, is_report=is_report)
        target_collection, target_tempfile = _getNetworkCollection(target_dir, is_report=is_report)
        finder = SubnetFinder.makeFromCombinations(
              reference_collection.networks,
              target_collection.networks,
              identity=identity,
              num_process=num_process)
        df = finder.find(is_report=is_report, max_num_assignment=max_num_mapping_pair)
    except:
        df = pd.DataFrame()
    finally:
        if reference_tempfile is not None:
            os.remove(reference_tempfile)
        if target_tempfile is not None:
            os.remove(target_tempfile)
    return df

def makeSerializationFile(directory:str, serialization_path:str, report_interval:int=10,
      is_report:bool=True)->None:
    """
    Serializes the models in the directory to a serialization file.

    Args:
        directory (str): Directory with Antimony models or SBML models
        serialization_path (str): Path to the serialization file to create (or overwrite)
        report_interval (int): Interval to report progress
        is_report (bool): If True, report progress
    """
    if not os.path.isdir(directory):
        raise ValueError(f"{directory} is not a directory")
    if not is_report:
        report_interval = int(1e6)
    #
    serializer = ModelSerializer(directory, serialization_path)
    serializer.serialize(report_interval=report_interval)


'''CompatibilityCollection consists of lists of rows in target that are compatible with a row in reference.'''


import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB import util # type: ignore

from memory_profiler import profile  # type: ignore
import itertools
import numpy as np
from typing import List, Tuple

NULL_NMAT = NamedMatrix(np.array([[]]))
NULL_INT = -1
NUM_SAMPLE = 1000  # Number of samples to estimate the number of assignments


class CompatibilityCollection(object):
    # A compatibility collection specifies the rows in self that are compatible with other.

    def __init__(self, num_self_row:int, num_other_row:int):
        self.num_self_row = num_self_row
        self.num_other_row = num_other_row
        self.compatibilities:list = [ [] for _ in range(num_self_row)]

    @classmethod
    def makeFromListOfLists(cls, list_of_lists:List[List[int]])->'CompatibilityCollection':
        num_self_row = len(list_of_lists)
        num_other_row = np.max([max(l) for l in list_of_lists])
        collection = CompatibilityCollection(num_self_row, num_other_row)
        collection.compatibilities = list_of_lists
        return collection

    def add(self, reference_row:int, target_rows:List[int]):
        # Add rows in target that are compatible with a row in reference
        self.compatibilities[reference_row].extend(target_rows)

    def copy(self)->'CompatibilityCollection':
        new_collection = CompatibilityCollection(self.num_self_row, self.num_other_row)
        new_collection.compatibilities = [l.copy() for l in self.compatibilities]
        return new_collection

    def __repr__(self):
        return str(self.compatibilities)
    
    def __len__(self)->int:
        return len(self.compatibilities)

    def __eq__(self, other)->bool:
        if len(self.compatibilities) != len(other.compatibilities):
            return False
        trues = [np.all(self.compatibilities[i] == other.compatibilities[i]) for i in range(len(self.compatibilities))]
        return bool(np.all(trues))
    
    def __le__(self, other)->bool:
        # self is less equal than other
        NUM_CHECK = 20
        # Allow for some variation because of random sampling of the compatibility collection
        comparisons = [other.log10_num_assignment - self.log10_num_assignment > -0.5
              for _ in range(NUM_CHECK)]
        return np.sum(comparisons) > NUM_CHECK*0.75

    @property
    def log10_num_assignment(self)->float:
        # Calculates the log of the number of assignments implied by the compatibility collection
        lengths = [len(l) for l in self.compatibilities]
        if 0 in lengths:
            return -np.inf
        unadjusted_estimate = np.sum([np.log10(len(l)) for l in self.compatibilities])
        # Estimate the number of assignments with duplicate indices
        sample_arr = util.sampleListOfLists(self.compatibilities, NUM_SAMPLE)
        unique_sample_arr = self._selectUnique(sample_arr)
        frac_unique = max(unique_sample_arr.shape[0] / sample_arr.shape[0], 1e-9)
        adjusted_estimate = max(0, unadjusted_estimate + np.log10(frac_unique))
        return adjusted_estimate

    @staticmethod
    def _selectUnique(array:np.ndarray)->np.ndarray:
        # Prunes rows in which there are duplicate values
        if len(array) > 0:
            idxs = np.all(np.diff(np.sort(array), axis=1) != 0, axis=1)
            result = array[idxs]
        else:
            result = np.array([])
        return result

    # FIXME: (1) Expand by batch. Do not merge batches. 
    # FIXME: (2) Make a generator
    #@profile
    def expand(self, max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT)->Tuple[np.ndarray, bool]:
        """Expands the compatibility collection into a two dimensional array of assignments

        Args:
            max_num_assignment (int, optional): maximum number of assignments Defaults to cn.MAX_NUM_ASSIGNMENT.

        Returns:
            np.ndarray: two dimensional array of assignments
            bool: is truncated
        """
        MAX_BATCH_SIZE = 1000
        #####
        def expandCollection(collection:List[List[int]])->np.ndarray:
            # Expands the compatibilities into a two dimensional array where each row is an assignment
            candidate_arr = np.array(list(itertools.product(*collection)))
            return self._selectUnique(candidate_arr)
        #####
        def mergeAssignments(assignment1:np.ndarray, assignment2:np.ndarray)->np.ndarray:
            # Merges two assignments into a single assignment
            # Check for degenerate conditions
            if assignment1.shape[0] == 0:
                return self._selectUnique(assignment2)
            if assignment2.shape[0] == 0:
                return self._selectUnique(assignment1)
            # Merge the assignments
            num_row1, num_row2 = assignment1.shape[0], assignment2.shape[0]
            big_assignment1 = np.repeat(assignment1, num_row2, axis=0)  # FIXME - big memory
            big_assignment2 = np.vstack([assignment2]*num_row1)
            merged_arr = np.concatenate([big_assignment1, big_assignment2], axis=1)  # FIXME - big memory
            result = self._selectUnique(merged_arr)  # FIXME - big memory
            return result
        #####
        # Form batches of assignments no more than the maximum size and then merge the batches
        log10_max_num_assignment = np.log10(max_num_assignment)
        collection, is_truncated = self.prune(log10_max_num_assignment=log10_max_num_assignment)
        list_of_lists = collection.compatibilities
        batches = []
        for _ in range(self.num_self_row):
            if len(list_of_lists) == 0:
                break
            # Construct a new batch
            num_assignment = 1
            assignment_idx = -1  # Index of the last assignment in the batch
            for idx in range(len(list_of_lists)):
                num_assignment *= len(list_of_lists[idx])
                if num_assignment > MAX_BATCH_SIZE:
                    break
                assignment_idx = idx
            #assert(assignment_idx >= 0)
            batch_collection = list_of_lists[0:assignment_idx+1]
            if assignment_idx < len(list_of_lists):
                list_of_lists = list_of_lists[assignment_idx+1:]
            else:
                list_of_lists = []
            expanded_collection = expandCollection(list(batch_collection))
            if expanded_collection.shape[0] == 0:
                return np.array([]), True
            batches.append(expandCollection(list(batch_collection)))
        # Estimate if the number of assignments is too large
        log10_estimate_num_assignment = np.sum([np.log10(len(l)) for l in batches])
        if log10_estimate_num_assignment > log10_max_num_assignment:
            return np.array([]), True
        # Merge the batches
        assignment_arr = batches[0]
        for idx in range(1, len(batches)):
            assignment_arr = mergeAssignments(assignment_arr, batches[idx])
        #
        #assert(assignment_arr.shape[1] == len(self.compatibilities))
        return assignment_arr, is_truncated

    def prune(self, log10_max_num_assignment:float)->Tuple['CompatibilityCollection', bool]:
        """Randomly prune the compatibility collection to a maximum number of assignments

        Args:
            log10_max_assignment (float): log10 of the maximum number of assignments

        Returns:
            CompatibilityCollection
            bool: is changed
        """
        collection = self.copy()
        #
        is_changed = False
        for idx in range(1000000):
            if collection.log10_num_assignment <= log10_max_num_assignment:
                break
            candidate_rows = [i for i in range(collection.num_self_row)
                              if len(collection.compatibilities[i]) > 1]  
            idx = np.random.randint(0, len(candidate_rows))
            irow = candidate_rows[idx]
            if len(collection.compatibilities[irow]) <= 1:
                continue
            # Check for duplicate single values
            pos = np.random.randint(0, len(collection.compatibilities[irow]))
            singles = list(np.array([v for v in collection.compatibilities if len(v) == 1]).flatten())
            lst = collection.compatibilities[irow][0:pos]
            lst.extend(collection.compatibilities[irow][pos+1:])
            if (len(lst) == 1) and (lst[0] in singles):
                continue
            # Delete the element
            del collection.compatibilities[irow][pos]
            is_changed = True
        else:
            raise ValueError("Could not prune the collection.")
        #
        return collection, is_changed